# Copyright 2019 Extreme Networks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from orquesta import conducting
from orquesta import events
from orquesta.specs import native as native_specs
from orquesta import statuses
from orquesta.tests.unit import base as test_base


class WorkflowConductorTaskRetryTest(test_base.WorkflowConductorTest):
    def test_basic_retry(self):
        wf_def = """
        version: 1.0

        tasks:
          task1:
            action: core.echo message="$RANDOM"
            retry:
              delay: 1
              count: 3
            next:
              - when: <% succeeded() %>
                do: task2
          task2:
            action: core.noop
        """

        expected_tk1_action_spec = {"action": "core.echo", "input": {"message": "$RANDOM"}}
        expected_tk2_action_spec = {"action": "core.noop", "input": None}

        spec = native_specs.WorkflowSpec(wf_def)
        self.assertDictEqual(spec.inspect(), {})

        conductor = conducting.WorkflowConductor(spec)
        conductor.request_workflow_status(statuses.RUNNING)

        # Failed execution for task1.
        next_tasks = conductor.get_next_tasks()
        self.assertEqual(len(next_tasks), 1)
        self.assertEqual(next_tasks[0]["id"], "task1")
        self.assertDictEqual(next_tasks[0]["actions"][0], expected_tk1_action_spec)
        self.forward_task_statuses(conductor, "task1", [statuses.RUNNING, statuses.FAILED])

        # Successful retry for task1.
        tk1_state = conductor.get_task_state_entry("task1", 0)
        self.assertEqual(tk1_state["status"], statuses.RETRYING)
        self.assertEqual(tk1_state["retry"]["count"], 3)
        self.assertEqual(tk1_state["retry"]["tally"], 1)
        next_tasks = conductor.get_next_tasks()
        self.assertEqual(len(next_tasks), 1)
        self.assertEqual(next_tasks[0]["id"], "task1")
        self.assertEqual(next_tasks[0]["delay"], 1)
        self.assertDictEqual(next_tasks[0]["actions"][0], expected_tk1_action_spec)
        self.forward_task_statuses(conductor, "task1", [statuses.RUNNING, statuses.SUCCEEDED])

        # Assert task1 succeeded and the workflow execution progresses to task2.
        tk1_state = conductor.get_task_state_entry("task1", 0)
        self.assertEqual(tk1_state["status"], statuses.SUCCEEDED)
        self.assertEqual(tk1_state["retry"]["tally"], 1)
        next_tasks = conductor.get_next_tasks()
        self.assertEqual(len(next_tasks), 1)
        self.assertEqual(next_tasks[0]["id"], "task2")
        self.assertDictEqual(next_tasks[0]["actions"][0], expected_tk2_action_spec)

        # Successful execution for task2.
        self.forward_task_statuses(conductor, "task2", [statuses.RUNNING, statuses.SUCCEEDED])
        tk2_state = conductor.get_task_state_entry("task2", 0)
        self.assertEqual(tk2_state["status"], statuses.SUCCEEDED)

        # Assert workflow completed successfully.
        self.assertEqual(conductor.get_workflow_status(), statuses.SUCCEEDED)

        # Assert there is only a single task1 in the task sequences.
        expected_task_sequence = ["task1", "task2"]
        actual_task_sequence = [item["id"] for item in conductor.workflow_state.sequence]
        self.assertListEqual(expected_task_sequence, actual_task_sequence)

    def test_multiple_tasks_with_retry(self):
        wf_def = """
        version: 1.0

        tasks:
          task1:
            action: core.echo message="$RANDOM"
            retry:
              delay: 1
              count: 3
            next:
              - when: <% succeeded() %>
                do: task3

          task2:
            action: core.echo message="$RANDOM"
            retry:
              delay: 1
              count: 3
            next:
              - when: <% succeeded() %>
                do: task4

          task3:
            action: core.noop

          task4:
            action: core.noop
        """

        expected_tk1_action_spec = {"action": "core.echo", "input": {"message": "$RANDOM"}}
        expected_tk2_action_spec = {"action": "core.echo", "input": {"message": "$RANDOM"}}
        expected_tk3_action_spec = {"action": "core.noop", "input": None}
        expected_tk4_action_spec = {"action": "core.noop", "input": None}

        spec = native_specs.WorkflowSpec(wf_def)
        self.assertDictEqual(spec.inspect(), {})

        conductor = conducting.WorkflowConductor(spec)
        conductor.request_workflow_status(statuses.RUNNING)

        # Failed execution for task1 and task2.
        next_tasks = conductor.get_next_tasks()
        self.assertEqual(len(next_tasks), 2)
        self.assertEqual(next_tasks[0]["id"], "task1")
        self.assertEqual(next_tasks[1]["id"], "task2")
        self.assertDictEqual(next_tasks[0]["actions"][0], expected_tk1_action_spec)
        self.forward_task_statuses(conductor, "task1", [statuses.RUNNING, statuses.FAILED])
        self.assertDictEqual(next_tasks[1]["actions"][0], expected_tk2_action_spec)
        self.forward_task_statuses(conductor, "task2", [statuses.RUNNING, statuses.FAILED])

        # Successful retry for task1.
        tk1_state = conductor.get_task_state_entry("task1", 0)
        self.assertEqual(tk1_state["status"], statuses.RETRYING)
        self.assertEqual(tk1_state["retry"]["count"], 3)
        self.assertEqual(tk1_state["retry"]["tally"], 1)
        next_tasks = conductor.get_next_tasks()
        self.assertEqual(len(next_tasks), 2)
        self.assertEqual(next_tasks[0]["id"], "task1")
        self.assertEqual(next_tasks[0]["delay"], 1)
        self.assertDictEqual(next_tasks[0]["actions"][0], expected_tk1_action_spec)
        self.forward_task_statuses(conductor, "task1", [statuses.RUNNING, statuses.SUCCEEDED])

        # Successful retry for task2.
        tk2_state = conductor.get_task_state_entry("task2", 0)
        self.assertEqual(tk2_state["status"], statuses.RETRYING)
        self.assertEqual(tk2_state["retry"]["count"], 3)
        self.assertEqual(tk2_state["retry"]["tally"], 1)
        next_tasks = conductor.get_next_tasks()
        self.assertEqual(len(next_tasks), 2)
        self.assertEqual(next_tasks[0]["id"], "task2")
        self.assertEqual(next_tasks[0]["delay"], 1)
        self.assertDictEqual(next_tasks[0]["actions"][0], expected_tk2_action_spec)
        self.forward_task_statuses(conductor, "task2", [statuses.RUNNING, statuses.SUCCEEDED])

        # Assert task1 succeeded and the workflow execution progresses to task3.
        tk1_state = conductor.get_task_state_entry("task1", 0)
        self.assertEqual(tk1_state["status"], statuses.SUCCEEDED)
        self.assertEqual(tk1_state["retry"]["tally"], 1)
        next_tasks = conductor.get_next_tasks()
        self.assertEqual(len(next_tasks), 2)
        self.assertEqual(next_tasks[0]["id"], "task3")
        self.assertDictEqual(next_tasks[0]["actions"][0], expected_tk3_action_spec)

        # Assert task2 succeeded and the workflow execution progresses to task4.
        tk2_state = conductor.get_task_state_entry("task2", 0)
        self.assertEqual(tk2_state["status"], statuses.SUCCEEDED)
        self.assertEqual(tk2_state["retry"]["tally"], 1)
        next_tasks = conductor.get_next_tasks()
        self.assertEqual(len(next_tasks), 2)
        self.assertEqual(next_tasks[1]["id"], "task4")
        self.assertDictEqual(next_tasks[1]["actions"][0], expected_tk4_action_spec)

        # Successful execution for task3.
        self.forward_task_statuses(conductor, "task3", [statuses.RUNNING, statuses.SUCCEEDED])
        tk3_state = conductor.get_task_state_entry("task3", 0)
        self.assertEqual(tk3_state["status"], statuses.SUCCEEDED)

        # Successful execution for task4.
        self.forward_task_statuses(conductor, "task4", [statuses.RUNNING, statuses.SUCCEEDED])
        tk4_state = conductor.get_task_state_entry("task4", 0)
        self.assertEqual(tk4_state["status"], statuses.SUCCEEDED)

        # Assert workflow completed successfully.
        self.assertEqual(conductor.get_workflow_status(), statuses.SUCCEEDED)

        # Assert there is only a single task1 in the task sequences.
        expected_task_sequence = ["task1", "task2", "task3", "task4"]
        actual_task_sequence = [item["id"] for item in conductor.workflow_state.sequence]
        self.assertListEqual(expected_task_sequence, actual_task_sequence)

    def test_retries_exhausted(self):
        wf_def = """
        version: 1.0

        tasks:
          task1:
            action: core.echo message="$RANDOM"
            retry:
              count: 3
            next:
              - when: <% succeeded() %>
                do: task2
          task2:
            action: core.noop
        """

        expected_tk1_action_spec = {"action": "core.echo", "input": {"message": "$RANDOM"}}

        spec = native_specs.WorkflowSpec(wf_def)
        self.assertDictEqual(spec.inspect(), {})

        conductor = conducting.WorkflowConductor(spec)
        conductor.request_workflow_status(statuses.RUNNING)

        # Failed execution for task1.
        next_tasks = conductor.get_next_tasks()
        self.assertEqual(len(next_tasks), 1)
        self.assertEqual(next_tasks[0]["id"], "task1")
        self.assertDictEqual(next_tasks[0]["actions"][0], expected_tk1_action_spec)
        self.forward_task_statuses(conductor, "task1", [statuses.RUNNING, statuses.FAILED])

        # Failed retry #1 for task1.
        tk1_state = conductor.get_task_state_entry("task1", 0)
        self.assertEqual(tk1_state["status"], statuses.RETRYING)
        self.assertEqual(tk1_state["retry"]["count"], 3)
        self.assertEqual(tk1_state["retry"]["tally"], 1)
        next_tasks = conductor.get_next_tasks()
        self.assertEqual(len(next_tasks), 1)
        self.assertEqual(next_tasks[0]["id"], "task1")
        self.assertDictEqual(next_tasks[0]["actions"][0], expected_tk1_action_spec)
        self.forward_task_statuses(conductor, "task1", [statuses.RUNNING, statuses.FAILED])

        # Failed retry #2 for task1.
        tk1_state = conductor.get_task_state_entry("task1", 0)
        self.assertEqual(tk1_state["status"], statuses.RETRYING)
        self.assertEqual(tk1_state["retry"]["tally"], 2)
        next_tasks = conductor.get_next_tasks()
        self.assertEqual(len(next_tasks), 1)
        self.assertEqual(next_tasks[0]["id"], "task1")
        self.assertDictEqual(next_tasks[0]["actions"][0], expected_tk1_action_spec)
        self.forward_task_statuses(conductor, "task1", [statuses.RUNNING, statuses.FAILED])

        # Failed retry #3 for task1.
        tk1_state = conductor.get_task_state_entry("task1", 0)
        self.assertEqual(tk1_state["status"], statuses.RETRYING)
        self.assertEqual(tk1_state["retry"]["tally"], 3)
        next_tasks = conductor.get_next_tasks()
        self.assertEqual(len(next_tasks), 1)
        self.assertEqual(next_tasks[0]["id"], "task1")
        self.assertDictEqual(next_tasks[0]["actions"][0], expected_tk1_action_spec)
        self.forward_task_statuses(conductor, "task1", [statuses.RUNNING, statuses.FAILED])

        # Assert task1 succeeded and the workflow execution progresses to task2.
        tk1_state = conductor.get_task_state_entry("task1", 0)
        self.assertEqual(tk1_state["status"], statuses.FAILED)
        self.assertEqual(tk1_state["retry"]["tally"], 3)
        next_tasks = conductor.get_next_tasks()
        self.assertEqual(len(next_tasks), 0)
        self.assertEqual(conductor.get_workflow_status(), statuses.FAILED)

        # Assert there is only a single task1 in the task sequences.
        expected_task_sequence = ["task1"]
        actual_task_sequence = [item["id"] for item in conductor.workflow_state.sequence]
        self.assertListEqual(expected_task_sequence, actual_task_sequence)

    def test_retries_exhausted_and_task_remediated(self):
        wf_def = """
        version: 1.0

        tasks:
          task1:
            action: core.echo message="$RANDOM"
            retry:
              count: 3
            next:
              - when: <% succeeded() %>
                do: task2
              - when: <% failed() %>
                do: task3
          task2:
            action: core.noop
          task3:
            action: core.echo message="BOOM!"
            next:
              - do: fail
        """

        expected_tk1_action_spec = {"action": "core.echo", "input": {"message": "$RANDOM"}}
        expected_tk3_action_spec = {"action": "core.echo", "input": {"message": "BOOM!"}}

        spec = native_specs.WorkflowSpec(wf_def)
        self.assertDictEqual(spec.inspect(), {})

        conductor = conducting.WorkflowConductor(spec)
        conductor.request_workflow_status(statuses.RUNNING)

        # Failed execution for task1.
        next_tasks = conductor.get_next_tasks()
        self.assertEqual(len(next_tasks), 1)
        self.assertEqual(next_tasks[0]["id"], "task1")
        self.assertDictEqual(next_tasks[0]["actions"][0], expected_tk1_action_spec)
        self.forward_task_statuses(conductor, "task1", [statuses.RUNNING, statuses.FAILED])

        # Failed retry #1 for task1.
        tk1_state = conductor.get_task_state_entry("task1", 0)
        self.assertEqual(tk1_state["status"], statuses.RETRYING)
        self.assertEqual(tk1_state["retry"]["count"], 3)
        self.assertEqual(tk1_state["retry"]["tally"], 1)
        next_tasks = conductor.get_next_tasks()
        self.assertEqual(len(next_tasks), 1)
        self.assertEqual(next_tasks[0]["id"], "task1")
        self.assertDictEqual(next_tasks[0]["actions"][0], expected_tk1_action_spec)
        self.forward_task_statuses(conductor, "task1", [statuses.RUNNING, statuses.FAILED])

        # Failed retry #2 for task1.
        tk1_state = conductor.get_task_state_entry("task1", 0)
        self.assertEqual(tk1_state["status"], statuses.RETRYING)
        self.assertEqual(tk1_state["retry"]["tally"], 2)
        next_tasks = conductor.get_next_tasks()
        self.assertEqual(len(next_tasks), 1)
        self.assertEqual(next_tasks[0]["id"], "task1")
        self.assertDictEqual(next_tasks[0]["actions"][0], expected_tk1_action_spec)
        self.forward_task_statuses(conductor, "task1", [statuses.RUNNING, statuses.FAILED])

        # Failed retry #3 for task1.
        tk1_state = conductor.get_task_state_entry("task1", 0)
        self.assertEqual(tk1_state["status"], statuses.RETRYING)
        self.assertEqual(tk1_state["retry"]["tally"], 3)
        next_tasks = conductor.get_next_tasks()
        self.assertEqual(len(next_tasks), 1)
        self.assertEqual(next_tasks[0]["id"], "task1")
        self.assertDictEqual(next_tasks[0]["actions"][0], expected_tk1_action_spec)
        self.forward_task_statuses(conductor, "task1", [statuses.RUNNING, statuses.FAILED])

        # Assert task1 failed and the workflow execution progresses to task3.
        tk1_state = conductor.get_task_state_entry("task1", 0)
        self.assertEqual(tk1_state["status"], statuses.FAILED)
        self.assertEqual(tk1_state["retry"]["tally"], 3)
        next_tasks = conductor.get_next_tasks()
        self.assertEqual(len(next_tasks), 1)
        self.assertEqual(next_tasks[0]["id"], "task3")
        self.assertDictEqual(next_tasks[0]["actions"][0], expected_tk3_action_spec)

        # Successful execution for task3.
        self.forward_task_statuses(conductor, "task3", [statuses.RUNNING, statuses.SUCCEEDED])
        tk3_state = conductor.get_task_state_entry("task3", 0)
        self.assertEqual(tk3_state["status"], statuses.SUCCEEDED)

        # Assert workflow failed (manual under task3).
        self.assertEqual(conductor.get_workflow_status(), statuses.FAILED)

        # Assert there is only a single task1 and a single task3 in the task sequences.
        expected_task_sequence = ["task1", "task3", "fail"]
        actual_task_sequence = [item["id"] for item in conductor.workflow_state.sequence]
        self.assertListEqual(expected_task_sequence, actual_task_sequence)

    def test_retry_delay_with_task_delay_defined(self):
        wf_def = """
        version: 1.0

        tasks:
          task1:
            delay: 2
            action: core.echo message="$RANDOM"
            retry:
              delay: 1
              count: 3
            next:
              - when: <% succeeded() %>
                do: task2
          task2:
            action: core.noop
        """

        expected_tk1_action_spec = {"action": "core.echo", "input": {"message": "$RANDOM"}}

        spec = native_specs.WorkflowSpec(wf_def)
        self.assertDictEqual(spec.inspect(), {})

        conductor = conducting.WorkflowConductor(spec)
        conductor.request_workflow_status(statuses.RUNNING)

        # Failed execution for task1.
        next_tasks = conductor.get_next_tasks()
        self.assertEqual(len(next_tasks), 1)
        self.assertEqual(next_tasks[0]["id"], "task1")
        self.assertEqual(next_tasks[0]["delay"], 2)
        self.assertDictEqual(next_tasks[0]["actions"][0], expected_tk1_action_spec)
        self.forward_task_statuses(conductor, "task1", [statuses.RUNNING, statuses.FAILED])

        # Successful retry for task1.
        tk1_state = conductor.get_task_state_entry("task1", 0)
        self.assertEqual(tk1_state["status"], statuses.RETRYING)
        self.assertEqual(tk1_state["retry"]["count"], 3)
        self.assertEqual(tk1_state["retry"]["tally"], 1)
        next_tasks = conductor.get_next_tasks()
        self.assertEqual(len(next_tasks), 1)
        self.assertEqual(next_tasks[0]["id"], "task1")
        self.assertEqual(next_tasks[0]["delay"], 1)
        self.assertDictEqual(next_tasks[0]["actions"][0], expected_tk1_action_spec)
        self.forward_task_statuses(conductor, "task1", [statuses.RUNNING, statuses.SUCCEEDED])

    def test_no_retry_delay_with_task_delay_defined(self):
        wf_def = """
        version: 1.0

        tasks:
          task1:
            delay: 2
            action: core.echo message="$RANDOM"
            retry:
              count: 3
            next:
              - when: <% succeeded() %>
                do: task2
          task2:
            action: core.noop
        """

        expected_tk1_action_spec = {"action": "core.echo", "input": {"message": "$RANDOM"}}

        spec = native_specs.WorkflowSpec(wf_def)
        self.assertDictEqual(spec.inspect(), {})

        conductor = conducting.WorkflowConductor(spec)
        conductor.request_workflow_status(statuses.RUNNING)

        # Failed execution for task1.
        next_tasks = conductor.get_next_tasks()
        self.assertEqual(len(next_tasks), 1)
        self.assertEqual(next_tasks[0]["id"], "task1")
        self.assertEqual(next_tasks[0]["delay"], 2)
        self.assertDictEqual(next_tasks[0]["actions"][0], expected_tk1_action_spec)
        self.forward_task_statuses(conductor, "task1", [statuses.RUNNING, statuses.FAILED])

        # Successful retry for task1.
        tk1_state = conductor.get_task_state_entry("task1", 0)
        self.assertEqual(tk1_state["status"], statuses.RETRYING)
        self.assertEqual(tk1_state["retry"]["count"], 3)
        self.assertEqual(tk1_state["retry"]["tally"], 1)
        next_tasks = conductor.get_next_tasks()
        self.assertEqual(len(next_tasks), 1)
        self.assertEqual(next_tasks[0]["id"], "task1")
        self.assertEqual(next_tasks[0]["delay"], 0)
        self.assertDictEqual(next_tasks[0]["actions"][0], expected_tk1_action_spec)
        self.forward_task_statuses(conductor, "task1", [statuses.RUNNING, statuses.SUCCEEDED])

    def test_workflow_cancellation_before_retry_running(self):
        wf_def = """
        version: 1.0

        tasks:
          task1:
            action: core.echo message="$RANDOM"
            retry:
              delay: 1
              count: 3
            next:
              - when: <% succeeded() %>
                do: task2
          task2:
            action: core.noop
        """

        expected_tk1_action_spec = {"action": "core.echo", "input": {"message": "$RANDOM"}}

        spec = native_specs.WorkflowSpec(wf_def)
        self.assertDictEqual(spec.inspect(), {})

        conductor = conducting.WorkflowConductor(spec)
        conductor.request_workflow_status(statuses.RUNNING)

        # Failed execution for task1.
        next_tasks = conductor.get_next_tasks()
        self.assertEqual(len(next_tasks), 1)
        self.assertEqual(next_tasks[0]["id"], "task1")
        self.assertDictEqual(next_tasks[0]["actions"][0], expected_tk1_action_spec)
        self.forward_task_statuses(conductor, "task1", [statuses.RUNNING, statuses.FAILED])
        tk1_state = conductor.get_task_state_entry("task1", 0)
        self.assertEqual(tk1_state["status"], statuses.RETRYING)

        # Request workflow cancellation.
        conductor.request_workflow_status(statuses.CANCELING)

        # Assert task1 is still retrying but not returned in get_next_tasks.
        tk1_state = conductor.get_task_state_entry("task1", 0)
        self.assertEqual(tk1_state["status"], statuses.RETRYING)
        next_tasks = conductor.get_next_tasks()
        self.assertEqual(len(next_tasks), 0)

        # Assert workflow is canceled.
        conductor.render_workflow_output()
        self.assertEqual(conductor.get_workflow_status(), statuses.CANCELED)

    def test_workflow_cancellation_while_retry_running(self):
        wf_def = """
        version: 1.0

        tasks:
          task1:
            action: core.echo message="$RANDOM"
            retry:
              delay: 1
              count: 3
            next:
              - when: <% succeeded() %>
                do: task2
          task2:
            action: core.noop
        """

        expected_tk1_action_spec = {"action": "core.echo", "input": {"message": "$RANDOM"}}

        spec = native_specs.WorkflowSpec(wf_def)
        self.assertDictEqual(spec.inspect(), {})

        conductor = conducting.WorkflowConductor(spec)
        conductor.request_workflow_status(statuses.RUNNING)

        # Failed execution for task1.
        next_tasks = conductor.get_next_tasks()
        self.assertEqual(len(next_tasks), 1)
        self.assertEqual(next_tasks[0]["id"], "task1")
        self.assertDictEqual(next_tasks[0]["actions"][0], expected_tk1_action_spec)
        self.forward_task_statuses(conductor, "task1", [statuses.RUNNING, statuses.FAILED])

        # Start retry for task1.
        tk1_state = conductor.get_task_state_entry("task1", 0)
        self.assertEqual(tk1_state["status"], statuses.RETRYING)
        self.assertEqual(tk1_state["retry"]["count"], 3)
        self.assertEqual(tk1_state["retry"]["tally"], 1)
        next_tasks = conductor.get_next_tasks()
        self.assertEqual(len(next_tasks), 1)
        self.assertEqual(next_tasks[0]["id"], "task1")
        self.assertEqual(next_tasks[0]["delay"], 1)
        self.assertDictEqual(next_tasks[0]["actions"][0], expected_tk1_action_spec)
        self.forward_task_statuses(conductor, "task1", [statuses.RUNNING])
        tk1_state = conductor.get_task_state_entry("task1", 0)
        self.assertEqual(tk1_state["status"], statuses.RUNNING)

        # Request workflow cancellation.
        conductor.request_workflow_status(statuses.CANCELING)
        self.assertEqual(conductor.get_workflow_status(), statuses.CANCELING)

        # Fail task1 and assert the task is not returned in get_next_tasks.
        self.forward_task_statuses(conductor, "task1", [statuses.FAILED])
        self.assertEqual(tk1_state["status"], statuses.RETRYING)
        next_tasks = conductor.get_next_tasks()
        self.assertEqual(len(next_tasks), 0)

        # Assert workflow is canceled.
        conductor.render_workflow_output()
        self.assertEqual(conductor.get_workflow_status(), statuses.CANCELED)

    def test_workflow_cancellation_with_multiple_retries_not_all_running(self):
        wf_def = """
        version: 1.0

        tasks:
          task1:
            action: core.echo message="$RANDOM"
            retry:
              delay: 1
              count: 3
            next:
              - when: <% succeeded() %>
                do: task3

          task2:
            action: core.echo message="$RANDOM"
            retry:
              delay: 1
              count: 3
            next:
              - when: <% succeeded() %>
                do: task4

          task3:
            action: core.noop

          task4:
            action: core.noop
        """

        spec = native_specs.WorkflowSpec(wf_def)
        self.assertDictEqual(spec.inspect(), {})

        conductor = conducting.WorkflowConductor(spec)
        conductor.request_workflow_status(statuses.RUNNING)

        # Failed execution for task1 and task2.
        self.forward_task_statuses(conductor, "task1", [statuses.RUNNING, statuses.FAILED])
        self.forward_task_statuses(conductor, "task2", [statuses.RUNNING, statuses.FAILED])

        # Start retry for task1.
        tk1_state = conductor.get_task_state_entry("task1", 0)
        self.assertEqual(tk1_state["status"], statuses.RETRYING)
        self.forward_task_statuses(conductor, "task1", [statuses.RUNNING])
        tk1_state = conductor.get_task_state_entry("task1", 0)
        self.assertEqual(tk1_state["status"], statuses.RUNNING)

        # Request workflow cancellation.
        conductor.request_workflow_status(statuses.CANCELING)
        self.assertEqual(conductor.get_workflow_status(), statuses.CANCELING)

        # Fail tasks and assert workflow transition from canceling to canceled.
        self.forward_task_statuses(conductor, "task1", [statuses.FAILED])
        tk1_state = conductor.get_task_state_entry("task1", 0)
        self.assertEqual(tk1_state["status"], statuses.RETRYING)
        tk2_state = conductor.get_task_state_entry("task2", 0)
        self.assertEqual(tk2_state["status"], statuses.RETRYING)
        self.assertEqual(conductor.get_workflow_status(), statuses.CANCELED)

    def test_workflow_cancellation_with_multiple_retries_running(self):
        wf_def = """
        version: 1.0

        tasks:
          task1:
            action: core.echo message="$RANDOM"
            retry:
              delay: 1
              count: 3
            next:
              - when: <% succeeded() %>
                do: task3

          task2:
            action: core.echo message="$RANDOM"
            retry:
              delay: 1
              count: 3
            next:
              - when: <% succeeded() %>
                do: task4

          task3:
            action: core.noop

          task4:
            action: core.noop
        """

        spec = native_specs.WorkflowSpec(wf_def)
        self.assertDictEqual(spec.inspect(), {})

        conductor = conducting.WorkflowConductor(spec)
        conductor.request_workflow_status(statuses.RUNNING)

        # Failed execution for task1 and task2.
        self.forward_task_statuses(conductor, "task1", [statuses.RUNNING, statuses.FAILED])
        self.forward_task_statuses(conductor, "task2", [statuses.RUNNING, statuses.FAILED])

        # Start retry for task1.
        tk1_state = conductor.get_task_state_entry("task1", 0)
        self.assertEqual(tk1_state["status"], statuses.RETRYING)
        self.forward_task_statuses(conductor, "task1", [statuses.RUNNING])
        tk1_state = conductor.get_task_state_entry("task1", 0)
        self.assertEqual(tk1_state["status"], statuses.RUNNING)

        # Start retry for task2.
        tk2_state = conductor.get_task_state_entry("task2", 0)
        self.assertEqual(tk2_state["status"], statuses.RETRYING)
        self.forward_task_statuses(conductor, "task2", [statuses.RUNNING])
        tk2_state = conductor.get_task_state_entry("task2", 0)
        self.assertEqual(tk2_state["status"], statuses.RUNNING)

        # Request workflow cancellation.
        conductor.request_workflow_status(statuses.CANCELING)
        self.assertEqual(conductor.get_workflow_status(), statuses.CANCELING)

        # Fail tasks and assert workflow transition from canceling to canceled.
        self.forward_task_statuses(conductor, "task1", [statuses.FAILED])
        tk1_state = conductor.get_task_state_entry("task1", 0)
        self.assertEqual(tk1_state["status"], statuses.RETRYING)
        self.assertEqual(conductor.get_workflow_status(), statuses.CANCELING)
        self.forward_task_statuses(conductor, "task2", [statuses.FAILED])
        tk2_state = conductor.get_task_state_entry("task2", 0)
        self.assertEqual(tk2_state["status"], statuses.RETRYING)
        self.assertEqual(conductor.get_workflow_status(), statuses.CANCELED)

    def test_workflow_pause_before_retry_running(self):
        wf_def = """
        version: 1.0

        tasks:
          task1:
            action: core.echo message="$RANDOM"
            retry:
              delay: 1
              count: 3
            next:
              - when: <% succeeded() %>
                do: task2
          task2:
            action: core.noop
        """

        expected_tk1_action_spec = {"action": "core.echo", "input": {"message": "$RANDOM"}}
        expected_tk2_action_spec = {"action": "core.noop", "input": None}

        spec = native_specs.WorkflowSpec(wf_def)
        self.assertDictEqual(spec.inspect(), {})

        conductor = conducting.WorkflowConductor(spec)
        conductor.request_workflow_status(statuses.RUNNING)

        # Failed execution for task1.
        next_tasks = conductor.get_next_tasks()
        self.assertEqual(len(next_tasks), 1)
        self.assertEqual(next_tasks[0]["id"], "task1")
        self.assertDictEqual(next_tasks[0]["actions"][0], expected_tk1_action_spec)
        self.forward_task_statuses(conductor, "task1", [statuses.RUNNING, statuses.FAILED])
        tk1_state = conductor.get_task_state_entry("task1", 0)
        self.assertEqual(tk1_state["status"], statuses.RETRYING)

        # Pause the workflow.
        conductor.request_workflow_status(statuses.PAUSING)

        # Assert task1 is still retrying but not returned in get_next_tasks.
        tk1_state = conductor.get_task_state_entry("task1", 0)
        self.assertEqual(tk1_state["status"], statuses.RETRYING)
        next_tasks = conductor.get_next_tasks()
        self.assertEqual(len(next_tasks), 0)

        # Assert workflow is canceled.
        conductor.render_workflow_output()
        self.assertEqual(conductor.get_workflow_status(), statuses.PAUSED)

        # Resume the workflow.
        conductor.request_workflow_status(statuses.RESUMING)

        # Successful retry for task1.
        tk1_state = conductor.get_task_state_entry("task1", 0)
        self.assertEqual(tk1_state["status"], statuses.RETRYING)
        self.assertEqual(tk1_state["retry"]["count"], 3)
        self.assertEqual(tk1_state["retry"]["tally"], 1)
        next_tasks = conductor.get_next_tasks()
        self.assertEqual(len(next_tasks), 1)
        self.assertEqual(next_tasks[0]["id"], "task1")
        self.assertEqual(next_tasks[0]["delay"], 1)
        self.assertDictEqual(next_tasks[0]["actions"][0], expected_tk1_action_spec)
        self.forward_task_statuses(conductor, "task1", [statuses.RUNNING, statuses.SUCCEEDED])

        # Assert task1 succeeded and the workflow execution progresses to task2.
        tk1_state = conductor.get_task_state_entry("task1", 0)
        self.assertEqual(tk1_state["status"], statuses.SUCCEEDED)
        self.assertEqual(tk1_state["retry"]["tally"], 1)
        next_tasks = conductor.get_next_tasks()
        self.assertEqual(len(next_tasks), 1)
        self.assertEqual(next_tasks[0]["id"], "task2")
        self.assertDictEqual(next_tasks[0]["actions"][0], expected_tk2_action_spec)

        # Successful execution for task2.
        self.forward_task_statuses(conductor, "task2", [statuses.RUNNING, statuses.SUCCEEDED])
        tk2_state = conductor.get_task_state_entry("task2", 0)
        self.assertEqual(tk2_state["status"], statuses.SUCCEEDED)

        # Assert workflow completed successfully.
        self.assertEqual(conductor.get_workflow_status(), statuses.SUCCEEDED)

        # Assert there is only a single task1 in the task sequences.
        expected_task_sequence = ["task1", "task2"]
        actual_task_sequence = [item["id"] for item in conductor.workflow_state.sequence]
        self.assertListEqual(expected_task_sequence, actual_task_sequence)

    def test_workflow_pause_while_retry_running(self):
        wf_def = """
        version: 1.0

        tasks:
          task1:
            action: core.echo message="$RANDOM"
            retry:
              delay: 1
              count: 3
            next:
              - when: <% succeeded() %>
                do: task2
          task2:
            action: core.noop
        """

        expected_tk1_action_spec = {"action": "core.echo", "input": {"message": "$RANDOM"}}
        expected_tk2_action_spec = {"action": "core.noop", "input": None}

        spec = native_specs.WorkflowSpec(wf_def)
        self.assertDictEqual(spec.inspect(), {})

        conductor = conducting.WorkflowConductor(spec)
        conductor.request_workflow_status(statuses.RUNNING)

        # Failed execution for task1.
        next_tasks = conductor.get_next_tasks()
        self.assertEqual(len(next_tasks), 1)
        self.assertEqual(next_tasks[0]["id"], "task1")
        self.assertDictEqual(next_tasks[0]["actions"][0], expected_tk1_action_spec)
        self.forward_task_statuses(conductor, "task1", [statuses.RUNNING, statuses.FAILED])

        # Start retry for task1.
        tk1_state = conductor.get_task_state_entry("task1", 0)
        self.assertEqual(tk1_state["status"], statuses.RETRYING)
        self.assertEqual(tk1_state["retry"]["count"], 3)
        self.assertEqual(tk1_state["retry"]["tally"], 1)
        next_tasks = conductor.get_next_tasks()
        self.assertEqual(len(next_tasks), 1)
        self.assertEqual(next_tasks[0]["id"], "task1")
        self.assertEqual(next_tasks[0]["delay"], 1)
        self.assertDictEqual(next_tasks[0]["actions"][0], expected_tk1_action_spec)
        self.forward_task_statuses(conductor, "task1", [statuses.RUNNING])
        tk1_state = conductor.get_task_state_entry("task1", 0)
        self.assertEqual(tk1_state["status"], statuses.RUNNING)

        # Pause the workflow.
        conductor.request_workflow_status(statuses.PAUSING)
        self.assertEqual(conductor.get_workflow_status(), statuses.PAUSING)

        # Fail task1 and assert the task is not returned in get_next_tasks.
        self.forward_task_statuses(conductor, "task1", [statuses.FAILED])
        self.assertEqual(tk1_state["status"], statuses.RETRYING)
        next_tasks = conductor.get_next_tasks()
        self.assertEqual(len(next_tasks), 0)

        # Assert the workflow is paused.
        self.assertEqual(conductor.get_workflow_status(), statuses.PAUSED)

        # Resume the workflow.
        conductor.request_workflow_status(statuses.RESUMING)

        # Successful retry for task1.
        tk1_state = conductor.get_task_state_entry("task1", 0)
        self.assertEqual(tk1_state["status"], statuses.RETRYING)
        self.assertEqual(tk1_state["retry"]["count"], 3)
        self.assertEqual(tk1_state["retry"]["tally"], 2)
        next_tasks = conductor.get_next_tasks()
        self.assertEqual(len(next_tasks), 1)
        self.assertEqual(next_tasks[0]["id"], "task1")
        self.assertEqual(next_tasks[0]["delay"], 1)
        self.assertDictEqual(next_tasks[0]["actions"][0], expected_tk1_action_spec)
        self.forward_task_statuses(conductor, "task1", [statuses.RUNNING, statuses.SUCCEEDED])

        # Assert task1 succeeded and the workflow execution progresses to task2.
        tk1_state = conductor.get_task_state_entry("task1", 0)
        self.assertEqual(tk1_state["status"], statuses.SUCCEEDED)
        self.assertEqual(tk1_state["retry"]["tally"], 2)
        next_tasks = conductor.get_next_tasks()
        self.assertEqual(len(next_tasks), 1)
        self.assertEqual(next_tasks[0]["id"], "task2")
        self.assertDictEqual(next_tasks[0]["actions"][0], expected_tk2_action_spec)

        # Successful execution for task2.
        self.forward_task_statuses(conductor, "task2", [statuses.RUNNING, statuses.SUCCEEDED])
        tk2_state = conductor.get_task_state_entry("task2", 0)
        self.assertEqual(tk2_state["status"], statuses.SUCCEEDED)

        # Assert workflow completed successfully.
        self.assertEqual(conductor.get_workflow_status(), statuses.SUCCEEDED)

        # Assert there is only a single task1 in the task sequences.
        expected_task_sequence = ["task1", "task2"]
        actual_task_sequence = [item["id"] for item in conductor.workflow_state.sequence]
        self.assertListEqual(expected_task_sequence, actual_task_sequence)

    def test_workflow_pause_with_multiple_retries_not_all_running(self):
        wf_def = """
        version: 1.0

        tasks:
          task1:
            action: core.echo message="$RANDOM"
            retry:
              delay: 1
              count: 3
            next:
              - when: <% succeeded() %>
                do: task3

          task2:
            action: core.echo message="$RANDOM"
            retry:
              delay: 1
              count: 3
            next:
              - when: <% succeeded() %>
                do: task4

          task3:
            action: core.noop

          task4:
            action: core.noop
        """

        spec = native_specs.WorkflowSpec(wf_def)
        self.assertDictEqual(spec.inspect(), {})

        conductor = conducting.WorkflowConductor(spec)
        conductor.request_workflow_status(statuses.RUNNING)

        # Failed execution for task1 and task2.
        self.forward_task_statuses(conductor, "task1", [statuses.RUNNING, statuses.FAILED])
        self.forward_task_statuses(conductor, "task2", [statuses.RUNNING, statuses.FAILED])

        # Start retry for task1.
        tk1_state = conductor.get_task_state_entry("task1", 0)
        self.assertEqual(tk1_state["status"], statuses.RETRYING)
        self.forward_task_statuses(conductor, "task1", [statuses.RUNNING])
        tk1_state = conductor.get_task_state_entry("task1", 0)
        self.assertEqual(tk1_state["status"], statuses.RUNNING)

        # Pause the workflow.
        conductor.request_workflow_status(statuses.PAUSING)
        self.assertEqual(conductor.get_workflow_status(), statuses.PAUSING)

        # Fail tasks and assert workflow transition from pausing to paused.
        self.forward_task_statuses(conductor, "task1", [statuses.FAILED])
        tk1_state = conductor.get_task_state_entry("task1", 0)
        self.assertEqual(tk1_state["status"], statuses.RETRYING)
        tk2_state = conductor.get_task_state_entry("task2", 0)
        self.assertEqual(tk2_state["status"], statuses.RETRYING)
        self.assertEqual(conductor.get_workflow_status(), statuses.PAUSED)

    def test_workflow_pause_with_multiple_retries_running(self):
        wf_def = """
        version: 1.0

        tasks:
          task1:
            action: core.echo message="$RANDOM"
            retry:
              delay: 1
              count: 3
            next:
              - when: <% succeeded() %>
                do: task3

          task2:
            action: core.echo message="$RANDOM"
            retry:
              delay: 1
              count: 3
            next:
              - when: <% succeeded() %>
                do: task4

          task3:
            action: core.noop

          task4:
            action: core.noop
        """

        spec = native_specs.WorkflowSpec(wf_def)
        self.assertDictEqual(spec.inspect(), {})

        conductor = conducting.WorkflowConductor(spec)
        conductor.request_workflow_status(statuses.RUNNING)

        # Failed execution for task1 and task2.
        self.forward_task_statuses(conductor, "task1", [statuses.RUNNING, statuses.FAILED])
        self.forward_task_statuses(conductor, "task2", [statuses.RUNNING, statuses.FAILED])

        # Start retry for task1.
        tk1_state = conductor.get_task_state_entry("task1", 0)
        self.assertEqual(tk1_state["status"], statuses.RETRYING)
        self.forward_task_statuses(conductor, "task1", [statuses.RUNNING])
        tk1_state = conductor.get_task_state_entry("task1", 0)
        self.assertEqual(tk1_state["status"], statuses.RUNNING)

        # Start retry for task2.
        tk2_state = conductor.get_task_state_entry("task2", 0)
        self.assertEqual(tk2_state["status"], statuses.RETRYING)
        self.forward_task_statuses(conductor, "task2", [statuses.RUNNING])
        tk2_state = conductor.get_task_state_entry("task2", 0)
        self.assertEqual(tk2_state["status"], statuses.RUNNING)

        # Pause the workflow.
        conductor.request_workflow_status(statuses.PAUSING)
        self.assertEqual(conductor.get_workflow_status(), statuses.PAUSING)

        # Fail tasks and assert workflow transition from pausing to paused.
        self.forward_task_statuses(conductor, "task1", [statuses.FAILED])
        tk1_state = conductor.get_task_state_entry("task1", 0)
        self.assertEqual(tk1_state["status"], statuses.RETRYING)
        self.assertEqual(conductor.get_workflow_status(), statuses.PAUSING)
        self.forward_task_statuses(conductor, "task2", [statuses.FAILED])
        tk2_state = conductor.get_task_state_entry("task2", 0)
        self.assertEqual(tk2_state["status"], statuses.RETRYING)
        self.assertEqual(conductor.get_workflow_status(), statuses.PAUSED)

    def test_retry_with_items(self):
        wf_def = """
        version: 1.0

        vars:
          - xs:
              - fee
              - fi
              - fo
              - fum

        tasks:
          task1:
            with: <% ctx(xs) %>
            action: core.echo message=<% item() %>
            retry:
              delay: 1
              count: 3
            next:
              - when: <% succeeded() %>
                do: task2

          task2:
            action: core.noop
        """

        result = [{"stdout": "fee"}, {"stdout": "fi"}, {"stdout": "fo"}, {"stdout": "fum"}]

        spec = native_specs.WorkflowSpec(wf_def)
        self.assertDictEqual(spec.inspect(), {})

        conductor = conducting.WorkflowConductor(spec)
        conductor.request_workflow_status(statuses.RUNNING)

        # Assert task1 is returned in get_next_tasks.
        next_tasks = conductor.get_next_tasks()
        self.assertEqual(len(next_tasks), 1)
        self.assertEqual(next_tasks[0]["id"], "task1")

        # Fail task1 and assert task1 is setup for retry.
        for i in range(0, 4):
            ac_ex_event = events.TaskItemActionExecutionEvent(i, statuses.RUNNING)
            conductor.update_task_state("task1", 0, ac_ex_event)

        for i in range(0, 4):
            ac_ex_event = events.TaskItemActionExecutionEvent(i, statuses.FAILED, None, result)
            conductor.update_task_state("task1", 0, ac_ex_event)

        # Assert task1 is setup for retry.
        tk1_state = conductor.get_task_state_entry("task1", 0)
        self.assertEqual(tk1_state["status"], statuses.RETRYING)
        next_tasks = conductor.get_next_tasks()
        self.assertEqual(len(next_tasks), 1)
        self.assertEqual(next_tasks[0]["id"], "task1")

        # Successful retry for task1.
        for i in range(0, 4):
            ac_ex_event = events.TaskItemActionExecutionEvent(i, statuses.RUNNING)
            conductor.update_task_state("task1", 0, ac_ex_event)

        for i in range(0, 4):
            ac_ex_event = events.TaskItemActionExecutionEvent(i, statuses.SUCCEEDED, None, result)
            conductor.update_task_state("task1", 0, ac_ex_event)

        # Assert task1 succeeded and the workflow execution progresses to task2.
        tk1_state = conductor.get_task_state_entry("task1", 0)
        self.assertEqual(tk1_state["status"], statuses.SUCCEEDED)
        self.assertEqual(tk1_state["retry"]["tally"], 1)
        next_tasks = conductor.get_next_tasks()
        self.assertEqual(len(next_tasks), 1)
        self.assertEqual(next_tasks[0]["id"], "task2")

        # Successful execution for task2.
        self.forward_task_statuses(conductor, "task2", [statuses.RUNNING, statuses.SUCCEEDED])
        tk2_state = conductor.get_task_state_entry("task2", 0)
        self.assertEqual(tk2_state["status"], statuses.SUCCEEDED)

        # Assert workflow completed successfully.
        self.assertEqual(conductor.get_workflow_status(), statuses.SUCCEEDED)

        # Assert there is only a single task1 in the task sequences.
        expected_task_sequence = ["task1", "task2"]
        actual_task_sequence = [item["id"] for item in conductor.workflow_state.sequence]
        self.assertListEqual(expected_task_sequence, actual_task_sequence)

    def test_workflow_cancellation_with_retry_running_with_items(self):
        wf_def = """
        version: 1.0

        vars:
          - xs:
              - fee
              - fi
              - fo
              - fum

        tasks:
          task1:
            with: <% ctx(xs) %>
            action: core.echo message=<% item() %>
            retry:
              delay: 1
              count: 3
            next:
              - when: <% succeeded() %>
                do: task2

          task2:
            action: core.noop
        """

        result = [{"stdout": "fee"}, {"stdout": "fi"}, {"stdout": "fo"}, {"stdout": "fum"}]

        spec = native_specs.WorkflowSpec(wf_def)
        self.assertDictEqual(spec.inspect(), {})

        conductor = conducting.WorkflowConductor(spec)
        conductor.request_workflow_status(statuses.RUNNING)

        # Assert task1 is returned in get_next_tasks.
        next_tasks = conductor.get_next_tasks()
        self.assertEqual(len(next_tasks), 1)
        self.assertEqual(next_tasks[0]["id"], "task1")

        # Fail task1 and assert task1 is setup for retry.
        for i in range(0, 4):
            ac_ex_event = events.TaskItemActionExecutionEvent(i, statuses.RUNNING)
            conductor.update_task_state("task1", 0, ac_ex_event)

        for i in range(0, 4):
            ac_ex_event = events.TaskItemActionExecutionEvent(i, statuses.FAILED, None, result)
            conductor.update_task_state("task1", 0, ac_ex_event)

        # Assert task1 is setup for retry.
        tk1_state = conductor.get_task_state_entry("task1", 0)
        self.assertEqual(tk1_state["status"], statuses.RETRYING)
        next_tasks = conductor.get_next_tasks()
        self.assertEqual(len(next_tasks), 1)
        self.assertEqual(next_tasks[0]["id"], "task1")

        # Start retry for task1.
        for i in range(0, 4):
            ac_ex_event = events.TaskItemActionExecutionEvent(i, statuses.RUNNING)
            conductor.update_task_state("task1", 0, ac_ex_event)

        # Request workflow cancellation.
        conductor.request_workflow_status(statuses.CANCELING)
        self.assertEqual(conductor.get_workflow_status(), statuses.CANCELING)

        # Fail the retry for task1.
        for i in range(0, 4):
            ac_ex_event = events.TaskItemActionExecutionEvent(i, statuses.FAILED)
            conductor.update_task_state("task1", 0, ac_ex_event)

        # Assert task1 is setup for retry.
        tk1_state = conductor.get_task_state_entry("task1", 0)
        self.assertEqual(tk1_state["status"], statuses.CANCELED)

        # Assert workflow is canceled.
        self.assertEqual(conductor.get_workflow_status(), statuses.CANCELED)

    def test_workflow_pause_with_retry_running_with_items(self):
        wf_def = """
        version: 1.0

        vars:
          - xs:
              - fee
              - fi
              - fo
              - fum

        tasks:
          task1:
            with: <% ctx(xs) %>
            action: core.echo message=<% item() %>
            retry:
              delay: 1
              count: 3
            next:
              - when: <% succeeded() %>
                do: task2

          task2:
            action: core.noop
        """

        result = [{"stdout": "fee"}, {"stdout": "fi"}, {"stdout": "fo"}, {"stdout": "fum"}]

        spec = native_specs.WorkflowSpec(wf_def)
        self.assertDictEqual(spec.inspect(), {})

        conductor = conducting.WorkflowConductor(spec)
        conductor.request_workflow_status(statuses.RUNNING)

        # Assert task1 is returned in get_next_tasks.
        next_tasks = conductor.get_next_tasks()
        self.assertEqual(len(next_tasks), 1)
        self.assertEqual(next_tasks[0]["id"], "task1")

        # Fail task1 and assert task1 is setup for retry.
        for i in range(0, 4):
            ac_ex_event = events.TaskItemActionExecutionEvent(i, statuses.RUNNING)
            conductor.update_task_state("task1", 0, ac_ex_event)

        for i in range(0, 4):
            ac_ex_event = events.TaskItemActionExecutionEvent(i, statuses.FAILED, None, result)
            conductor.update_task_state("task1", 0, ac_ex_event)

        # Assert task1 is setup for retry.
        tk1_state = conductor.get_task_state_entry("task1", 0)
        self.assertEqual(tk1_state["status"], statuses.RETRYING)
        next_tasks = conductor.get_next_tasks()
        self.assertEqual(len(next_tasks), 1)
        self.assertEqual(next_tasks[0]["id"], "task1")

        # Start retry for task1.
        for i in range(0, 4):
            ac_ex_event = events.TaskItemActionExecutionEvent(i, statuses.RUNNING)
            conductor.update_task_state("task1", 0, ac_ex_event)

        # Pause workflow.
        conductor.request_workflow_status(statuses.PAUSING)
        self.assertEqual(conductor.get_workflow_status(), statuses.PAUSING)

        # Succeed the retry for task1.
        for i in range(0, 4):
            ac_ex_event = events.TaskItemActionExecutionEvent(i, statuses.FAILED)
            conductor.update_task_state("task1", 0, ac_ex_event)

        # Assert task1 is setup for retry.
        tk1_state = conductor.get_task_state_entry("task1", 0)
        self.assertEqual(tk1_state["status"], statuses.RETRYING)

        # Assert workflow is canceled.
        self.assertEqual(conductor.get_workflow_status(), statuses.PAUSED)

    def test_bad_task_name(self):
        wf_def = """
        version: 1.0

        tasks:
          retry:
            action: core.echo message="$RANDOM"
            next:
              - when: <% succeeded() %>
                do: task2
          task2:
            action: core.noop
        """

        expected_inspection_errors = {
            "semantics": [
                {
                    "message": 'The task name "retry" is reserved with special function.',
                    "schema_path": "properties.tasks.patternProperties.^\\w+$",
                    "spec_path": "tasks.retry",
                }
            ]
        }

        spec = native_specs.WorkflowSpec(wf_def)

        self.assertDictEqual(spec.inspect(), expected_inspection_errors)

    def test_retry_command(self):
        wf_def = """
        version: 1.0

        tasks:
          task1:
            action: core.echo message="$RANDOM"
            next:
              - when: <% failed() %>
                do: retry
              - when: <% succeeded() %>
                do: task2
          task2:
            action: core.noop
        """

        expected_tk1_action_spec = {"action": "core.echo", "input": {"message": "$RANDOM"}}
        expected_tk2_action_spec = {"action": "core.noop", "input": None}

        spec = native_specs.WorkflowSpec(wf_def)
        self.assertDictEqual(spec.inspect(), {})

        conductor = conducting.WorkflowConductor(spec)
        conductor.request_workflow_status(statuses.RUNNING)

        # Failed execution for task1.
        next_tasks = conductor.get_next_tasks()
        self.assertEqual(len(next_tasks), 1)
        self.assertEqual(next_tasks[0]["id"], "task1")
        self.forward_task_statuses(conductor, "task1", [statuses.RUNNING, statuses.FAILED])

        # Successful retry for task1.
        tk1_state = conductor.get_task_state_entry("task1", 0)
        self.assertEqual(tk1_state["status"], statuses.RETRYING)
        self.assertEqual(tk1_state["retry"]["count"], 3)
        self.assertEqual(tk1_state["retry"]["tally"], 1)
        next_tasks = conductor.get_next_tasks()
        self.assertEqual(len(next_tasks), 1)
        self.assertEqual(next_tasks[0]["id"], "task1")
        self.assertEqual(next_tasks[0]["delay"], 0)
        self.assertDictEqual(next_tasks[0]["actions"][0], expected_tk1_action_spec)
        self.forward_task_statuses(conductor, "task1", [statuses.RUNNING, statuses.SUCCEEDED])

        # Assert task1 succeeded and the workflow execution progresses to task2.
        tk1_state = conductor.get_task_state_entry("task1", 0)
        self.assertEqual(tk1_state["status"], statuses.SUCCEEDED)
        self.assertEqual(tk1_state["retry"]["tally"], 1)
        next_tasks = conductor.get_next_tasks()
        self.assertEqual(len(next_tasks), 1)
        self.assertEqual(next_tasks[0]["id"], "task2")
        self.assertDictEqual(next_tasks[0]["actions"][0], expected_tk2_action_spec)

        # Successful execution for task2.
        self.forward_task_statuses(conductor, "task2", [statuses.RUNNING, statuses.SUCCEEDED])
        tk2_state = conductor.get_task_state_entry("task2", 0)
        self.assertEqual(tk2_state["status"], statuses.SUCCEEDED)

        # Assert workflow completed successfully.
        self.assertEqual(conductor.get_workflow_status(), statuses.SUCCEEDED)

        # Assert there is only a single task1 in the task sequences.
        expected_task_sequence = ["task1", "task2"]
        actual_task_sequence = [item["id"] for item in conductor.workflow_state.sequence]
        self.assertListEqual(expected_task_sequence, actual_task_sequence)



#!/usr/bin/env python2.7

# Copyright 2019 Extreme Networks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import sys

from setuptools import setup, find_packages


PKG_ROOT_DIR = os.path.dirname(os.path.realpath(__file__))
PKG_REQ_FILE = "%s/requirements.txt" % PKG_ROOT_DIR
os.chdir(PKG_ROOT_DIR)


def get_version_string():
    version = None
    sys.path.insert(0, PKG_ROOT_DIR)
    from orquesta import __version__

    version = __version__
    sys.path.pop(0)
    return version


def get_requirements():
    with open(PKG_REQ_FILE) as f:
        required = f.read().splitlines()

    # Ignore comments in the requirements file
    required = [line for line in required if not line.startswith("#")]
    return required


setup(
    name="orquesta",
    version=get_version_string(),
    author="StackStorm",
    author_email="info@stackstorm.com",
    url="https://www.stackstorm.com",
    packages=find_packages(exclude=[]),
    install_requires=get_requirements(),
    license="Apache License (2.0)",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Information Technology",
        "Intended Audience :: System Administrators",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: POSIX :: Linux",
        "Programming Language :: Python",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
    ],
    entry_points={
        "orquesta.composers": [
            "native = orquesta.composers.native:WorkflowComposer",
            "mistral = orquesta.composers.mistral:WorkflowComposer",
            "mock = orquesta.composers.mock:WorkflowComposer",
        ],
        "orquesta.expressions.evaluators": [
            "yaql = orquesta.expressions.yql:YAQLEvaluator",
            "jinja = orquesta.expressions.jinja:JinjaEvaluator",
        ],
        "orquesta.expressions.functions": [
            "ctx = orquesta.expressions.functions.common:ctx_",
            "json = orquesta.expressions.functions.common:json_",
            "zip = orquesta.expressions.functions.common:zip_",
            "item = orquesta.expressions.functions.workflow:item_",
            "task_status = orquesta.expressions.functions.workflow:task_status_",
            "succeeded = orquesta.expressions.functions.workflow:succeeded_",
            "failed = orquesta.expressions.functions.workflow:failed_",
            "completed = orquesta.expressions.functions.workflow:completed_",
            "result = orquesta.expressions.functions.workflow:result_",
        ],
        "orquesta.tests": ["fake = orquesta.tests.unit.utils.test_plugin:FakePlugin"],
    },
)



# Copyright 2019 Extreme Networks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from orquesta.expressions import base as expr_base
from orquesta.tests.unit import base as test_base


class JinjaFacadeVariableExtractionTest(test_base.ExpressionFacadeEvaluatorTest):
    def test_empty_extraction(self):
        expr = (
            "{{ just_text and $not_a_var and notctx().bar and "
            "ctx(). and ctx().() and ctx().-foobar and ctx().foobar() }}"
        )

        self.assertListEqual([], expr_base.extract_vars(expr))

    def test_single_var_extraction(self):
        expr = "{{ ctx().foobar }}"

        expected_vars = [("jinja", expr, "foobar")]

        self.assertListEqual(expected_vars, expr_base.extract_vars(expr))

    def test_single_dotted_var_extraction(self):
        expr = "{{ ctx().foo.bar }}"

        expected_vars = [("jinja", expr, "foo")]

        self.assertListEqual(expected_vars, expr_base.extract_vars(expr))

    def test_single_indexing_var_extraction(self):
        expr = "{{ ctx().foo[0] }}"

        expected_vars = [("jinja", expr, "foo")]

        self.assertListEqual(expected_vars, expr_base.extract_vars(expr))

    def test_single_functional_var_extraction(self):
        expr = "{{ ctx().foo.get(bar) }}"

        expected_vars = [("jinja", expr, "foo")]

        self.assertListEqual(expected_vars, expr_base.extract_vars(expr))

    def test_multiple_vars_extraction(self):
        expr = "{{ctx().fubar ctx().foobar ctx().foo.get(bar) ctx().fu.bar ctx().foobaz.bar[0] }}"

        expected_vars = [
            ("jinja", expr, "foo"),
            ("jinja", expr, "foobar"),
            ("jinja", expr, "foobaz"),
            ("jinja", expr, "fu"),
            ("jinja", expr, "fubar"),
        ]

        self.assertListEqual(expected_vars, expr_base.extract_vars(expr))

    def test_multiple_interleaved_vars_extraction(self):
        expr = "{{ Why the ctx().foobar are you so ctx().fu.bar serious? }}"

        expected_vars = [("jinja", expr, "foobar"), ("jinja", expr, "fu")]

        self.assertListEqual(expected_vars, expr_base.extract_vars(expr))

    def test_vars_extraction_from_list(self):
        expr = [
            "{{ abc }}",
            "{{ Why the ctx().foobar are you so ctx().fu.bar serious? }}",
            "All your test_base are belong to us.",
            {"{{ ctx().x }}": 123, "k2": "{{ ctx().y }}", "k3": ["{{ ctx().z }}"]},
        ]

        expected_vars = [
            ("jinja", expr[1], "foobar"),
            ("jinja", expr[1], "fu"),
            ("jinja", "{{ ctx().x }}", "x"),
            ("jinja", "{{ ctx().y }}", "y"),
            ("jinja", "{{ ctx().z }}", "z"),
        ]

        self.assertListEqual(expected_vars, expr_base.extract_vars(expr))

    def test_vars_extraction_from_dict(self):
        expr = {
            "k1": "{{ abc }}",
            "k2": "{{ Why the ctx().foobar are you so ctx().fu.bar serious? }}",
            "k3": ["{{ ctx().z }}"],
            "k4": {"k5": "{{ ctx().y }}"},
            "{{ ctx().x }}": 123,
        }

        expected_vars = [
            ("jinja", expr["k2"], "foobar"),
            ("jinja", expr["k2"], "fu"),
            ("jinja", "{{ ctx().x }}", "x"),
            ("jinja", "{{ ctx().y }}", "y"),
            ("jinja", "{{ ctx().z }}", "z"),
        ]

        self.assertListEqual(expected_vars, expr_base.extract_vars(expr))



# Copyright 2019 Extreme Networks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Use copy.deepcopy instead of orquesta.utils.jsonify.deepcopy because
# the schema contains instance(s) of Spec class(es) in which the
# jsonify.deepcopy method will convert the Spec class(es) to dict.
import copy

from orquesta import exceptions as exc
from orquesta.utils import dictionary as dict_util


def get_schema_type(s):
    return (s or {}).get("type")


def check_schema_mergeable(s):
    schema_type = get_schema_type(s)

    if s and schema_type not in ["array", "object"]:
        raise exc.SchemaIncompatibleError(
            "Merge operation is only supported for schema of type array "
            "and object. The given schema is type %s." % str(schema_type)
        )


def check_schemas_compatible(s1, s2):
    s1_type = get_schema_type(s1)
    s2_type = get_schema_type(s2)

    if s1_type != s2_type:
        raise exc.SchemaIncompatibleError(
            "Merge requires both types of schema to be the same. "
            "The schema types are %s and %s." % (s1_type, s2_type)
        )

    return s1_type


def merge_schema(s1, s2, overwrite=True):
    blank_schema_templates = [{"type": "object"}, {"type": "array"}]

    if s1 and s1 not in blank_schema_templates and not s2:
        return copy.deepcopy(s1)

    if s1 and s1 in blank_schema_templates and not s2:
        return {}

    if not s1 and s2 and s2 not in blank_schema_templates:
        return copy.deepcopy(s2)

    if not s1 and s2 and s2 in blank_schema_templates:
        return {}

    if not s1 and not s2:
        return {}

    if s1 in blank_schema_templates and s2 not in blank_schema_templates:
        return copy.deepcopy(s2)

    if s1 not in blank_schema_templates and s2 in blank_schema_templates:
        return copy.deepcopy(s1)

    if s1 in blank_schema_templates and s2 in blank_schema_templates:
        return {}

    check_schema_mergeable(s1)
    check_schema_mergeable(s2)
    schema_type = check_schemas_compatible(s1, s2)
    schema = SCHEMA_MERGE_FUNCTIONS[schema_type](s1, s2, overwrite=overwrite)

    return schema


def merge_object_schema(s1, s2, overwrite=True):
    schema = {"type": "object"}

    properties = dict_util.merge_dicts(
        copy.deepcopy(s1.get("properties", {})),
        copy.deepcopy(s2.get("properties", {})),
        overwrite=overwrite,
    )

    if properties:
        schema["properties"] = properties

    required = list(
        set(copy.deepcopy(s1.get("required", []))).union(set(copy.deepcopy(s2.get("required", []))))
    )

    if required:
        schema["required"] = sorted(required)

    additional = s1.get("additionalProperties", True) and s2.get("additionalProperties", True)

    if not additional:
        schema["additionalProperties"] = additional

    pattern_properties = dict_util.merge_dicts(
        copy.deepcopy(s1.get("patternProperties", {})),
        copy.deepcopy(s2.get("patternProperties", {})),
        overwrite=overwrite,
    )

    if pattern_properties:
        schema["patternProperties"] = pattern_properties

    min_properties = (
        s1.get("minProperties", 0)
        if not overwrite
        else max(s1.get("minProperties", 0), s2.get("minProperties", 0))
    )

    if min_properties > 0:
        schema["minProperties"] = min_properties

    max_properties = (
        s1.get("maxProperties", 0)
        if not overwrite
        else min(s1.get("maxProperties", 0), s2.get("maxProperties", 0))
    )

    if max_properties > 0:
        schema["maxProperties"] = max_properties

    return schema


def merge_array_schema(s1, s2, overwrite=True):
    schema = {"type": "array"}

    def process_schema_property(name, default):
        return s2.get(name, default) if overwrite else s1.get(name, default)

    items = process_schema_property("items", [])

    if items:
        schema["items"] = items

    unique_items = process_schema_property("uniqueItems", False)

    if unique_items:
        schema["uniqueItems"] = unique_items

    min_items = process_schema_property("minItems", 0)

    if min_items > 0:
        schema["minItems"] = min_items

    max_items = process_schema_property("maxItems", 0)

    if max_items > 0:
        schema["maxItems"] = max_items

    return schema


SCHEMA_MERGE_FUNCTIONS = {"object": merge_object_schema, "array": merge_array_schema}



# Copyright 2019 Extreme Networks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest

from orquesta.utils import context as ctx_util
from orquesta.utils import jsonify as json_util


class ContextUtilTest(unittest.TestCase):
    def test_set_current_task(self):
        context = {"var1": "foobar"}
        task = {"id": "t1", "route": 0}

        context = ctx_util.set_current_task(context, task)
        expected_context = dict(
            [("__current_task", json_util.deepcopy(task))] + list(context.items())
        )

        self.assertDictEqual(context, expected_context)

    def test_set_current_task_with_result(self):
        context = {"var1": "foobar"}
        task = {"id": "t1", "route": 0, "result": "foobar"}

        context = ctx_util.set_current_task(context, task)
        expected_context = dict(
            [("__current_task", json_util.deepcopy(task))] + list(context.items())
        )

        self.assertDictEqual(context, expected_context)

    def test_set_current_task_nonetype_context(self):
        task = {"id": "t1", "route": 0}

        context = ctx_util.set_current_task(None, task)
        expected_context = {"__current_task": json_util.deepcopy(task)}

        self.assertDictEqual(context, expected_context)

    def test_set_current_task_empty_context(self):
        task = {"id": "t1", "route": 0}

        context = ctx_util.set_current_task(dict(), task)
        expected_context = {"__current_task": json_util.deepcopy(task)}

        self.assertDictEqual(context, expected_context)

    def test_set_current_task_empty_task(self):
        context = {"var1": "foobar"}

        self.assertRaises(ValueError, ctx_util.set_current_task, context, dict())

    def test_set_current_task_bad_types(self):
        task = {"id": "t1", "route": 0}

        self.assertRaises(TypeError, ctx_util.set_current_task, "foobar", task)

        self.assertRaises(TypeError, ctx_util.set_current_task, dict(), "foobar")



# Copyright 2019 Extreme Networks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from orquesta.expressions import yql as yaql_expr
from orquesta.tests.unit import base as test_base
from orquesta.utils import plugin as plugin_util


class YAQLValidationTest(test_base.ExpressionEvaluatorTest):
    @classmethod
    def setUpClass(cls):
        cls.language = "yaql"
        super(YAQLValidationTest, cls).setUpClass()

    def test_get_evaluator(self):
        self.assertEqual(
            plugin_util.get_module("orquesta.expressions.evaluators", self.language),
            yaql_expr.YAQLEvaluator,
        )

    def test_basic_validate(self):
        self.assertListEqual([], self.evaluator.validate("<% 1 %>"))
        self.assertListEqual([], self.evaluator.validate("<% abc %>"))
        self.assertListEqual([], self.evaluator.validate("<% 1 + 2 %>"))
        self.assertListEqual([], self.evaluator.validate("<% ctx().foo %>"))
        self.assertListEqual([], self.evaluator.validate("<% ctx(foo) %>"))
        self.assertListEqual([], self.evaluator.validate("<% ctx().a1 + ctx(a2) %>"))

    def test_parse_error(self):
        expr = "<% <% ctx().foo %> %>"
        errors = self.evaluator.validate(expr)

        self.assertEqual(1, len(errors))
        self.assertIn("Parse error", errors[0]["message"])

    def test_lexical_error(self):
        expr = '<% {"a": 123} %>'
        errors = self.evaluator.validate(expr)

        self.assertEqual(1, len(errors))
        self.assertIn("Lexical error", errors[0]["message"])

    def test_multiple_errors(self):
        expr = '<% 1 +/ 2 %> and <% {"a": 123} %>'
        errors = self.evaluator.validate(expr)

        self.assertEqual(2, len(errors))
        self.assertIn("Parse error", errors[0]["message"])
        self.assertIn("Lexical error", errors[1]["message"])



# Copyright 2019 Extreme Networks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest

from orquesta.utils import dictionary as dict_util
from orquesta.utils import jsonify as json_util


LEFT = {"k1": "123", "k2": "abc", "k3": {"k31": True, "k32": 1.0}}

RIGHT = {"k2": "def", "k3": {"k32": 2.0, "k33": {"k331": "foo"}}, "k4": "bar"}


class DictUtilsTest(unittest.TestCase):
    def test_dict_merge_overwrite(self):
        left = json_util.deepcopy(LEFT)
        right = json_util.deepcopy(RIGHT)

        dict_util.merge_dicts(left, right)

        expected = {
            "k1": "123",
            "k2": "def",
            "k3": {"k31": True, "k32": 2.0, "k33": {"k331": "foo"}},
            "k4": "bar",
        }

        self.assertDictEqual(left, expected)

    def test_dict_merge_overwrite_false(self):
        left = json_util.deepcopy(LEFT)
        right = json_util.deepcopy(RIGHT)

        dict_util.merge_dicts(left, right, overwrite=False)

        expected = {
            "k1": "123",
            "k2": "abc",
            "k3": {"k31": True, "k32": 1.0, "k33": {"k331": "foo"}},
            "k4": "bar",
        }

        self.assertDictEqual(left, expected)

    def test_dict_dot_notation_access(self):
        data = {
            "a": "foo",
            "b": {"c": "bar", "d": {"e": 123, "f": False, "g": {}, "h": None}},
            "x": {},
            "y": None,
        }

        self.assertEqual("foo", dict_util.get_dict_value(data, "a"))
        self.assertEqual("bar", dict_util.get_dict_value(data, "b.c"))
        self.assertEqual(123, dict_util.get_dict_value(data, "b.d.e"))
        self.assertFalse(dict_util.get_dict_value(data, "b.d.f"))
        self.assertDictEqual({}, dict_util.get_dict_value(data, "b.d.g"))
        self.assertIsNone(dict_util.get_dict_value(data, "b.d.h"))
        self.assertDictEqual({}, dict_util.get_dict_value(data, "x"))
        self.assertIsNone(dict_util.get_dict_value(data, "x.x"))
        self.assertIsNone(dict_util.get_dict_value(data, "y"))
        self.assertIsNone(dict_util.get_dict_value(data, "z"))

    def test_dict_dot_notation_access_type_error(self):
        data = {"a": "foo"}

        self.assertRaises(TypeError, dict_util.get_dict_value, data, "a.b")

    def test_dict_dot_notation_access_key_error(self):
        data = {"a": {}}

        self.assertRaises(KeyError, dict_util.get_dict_value, data, "a.b", raise_key_error=True)

    def test_dict_dot_notation_set_value(self):
        data = {
            "a": "foo",
            "b": {"c": "bar", "d": {"e": 123, "f": False, "g": {}, "h": None}},
            "x": {},
            "y": None,
        }

        # Test basic insert.
        dict_util.set_dict_value(data, "z", {"foo": "bar"})

        # Test insert via dot notation on existing node.
        dict_util.set_dict_value(data, "b.d.h", 2.0)

        # Test insert via dot notation on nonexistent nodes.
        dict_util.set_dict_value(data, "m.n.o", True)

        # Test insert non-null only.
        dict_util.set_dict_value(data, "b.d.i", None, insert_null=False)

        # Test insert null.
        dict_util.set_dict_value(data, "b.d.j", None, insert_null=True)

        expected_data = {
            "a": "foo",
            "b": {"c": "bar", "d": {"e": 123, "f": False, "g": {}, "h": 2.0, "j": None}},
            "m": {"n": {"o": True}},
            "x": {},
            "y": None,
            "z": {"foo": "bar"},
        }

        self.assertDictEqual(data, expected_data)

    def test_dict_dot_notation_set_value_type_error(self):
        data = {"a": "foo"}

        self.assertRaises(TypeError, dict_util.set_dict_value, data, "a.b", "foobar")

    def test_dict_dot_notation_set_value_key_error(self):
        data = {"a": {}}

        self.assertRaises(
            KeyError, dict_util.set_dict_value, data, "a.b", "foobar", raise_key_error=True
        )



# Copyright 2019 Extreme Networks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import abc
import inspect
import logging
import re
import six
import threading

from stevedore import extension

from orquesta.utils import expression as expr_util
from orquesta.utils import plugin as plugin_util


LOG = logging.getLogger(__name__)

_EXP_EVALUATORS = None
_EXP_EVALUATORS_LOCK = threading.Lock()
_EXP_EVALUATOR_NAMESPACE = "orquesta.expressions.evaluators"


@six.add_metaclass(abc.ABCMeta)
class Evaluator(object):
    _type = "unspecified"
    _delimiter = None

    @classmethod
    def get_type(cls):
        return cls._type

    @classmethod
    def strip_delimiter(cls, expr):
        return expr.strip(cls._delimiter).strip()

    @classmethod
    def get_statement_regex(cls):
        raise NotImplementedError()

    @classmethod
    def has_expressions(cls, text):
        raise NotImplementedError()

    @classmethod
    @abc.abstractmethod
    def validate(cls, statement):
        raise NotImplementedError()

    @classmethod
    @abc.abstractmethod
    def evaluate(cls, text, data=None):
        raise NotImplementedError()

    @classmethod
    @abc.abstractmethod
    def extract_vars(cls, statement):
        raise NotImplementedError()


def get_evaluator(language):
    return plugin_util.get_module(_EXP_EVALUATOR_NAMESPACE, language)


def get_evaluators():
    global _EXP_EVALUATORS
    global _EXP_EVALUATORS_LOCK

    with _EXP_EVALUATORS_LOCK:
        if _EXP_EVALUATORS is None:
            _EXP_EVALUATORS = {}

            mgr = extension.ExtensionManager(
                namespace=_EXP_EVALUATOR_NAMESPACE, invoke_on_load=False
            )

            for name in mgr.names():
                _EXP_EVALUATORS[name] = get_evaluator(name)

    return _EXP_EVALUATORS


def get_statement_regexes():
    return {t: e.get_statement_regex() for t, e in six.iteritems(get_evaluators())}


def has_expressions(text):
    result = {t: e.has_expressions(text) for t, e in six.iteritems(get_evaluators())}

    return any(result.values())


def validate(statement):
    errors = []

    if isinstance(statement, dict):
        for k, v in six.iteritems(statement):
            errors.extend(validate(k)["errors"])
            errors.extend(validate(v)["errors"])

    elif isinstance(statement, list):
        for item in statement:
            errors.extend(validate(item)["errors"])

    elif isinstance(statement, six.string_types):
        evaluators = [
            evaluator
            for name, evaluator in six.iteritems(get_evaluators())
            if evaluator.has_expressions(statement)
        ]

        if len(evaluators) == 1:
            errors.extend(evaluators[0].validate(statement))
        elif len(evaluators) > 1:
            message = "Expression with multiple types is not supported."
            errors.append(expr_util.format_error(None, statement, message))

    return {"errors": errors}


def evaluate(statement, data=None):
    if isinstance(statement, dict):
        return {evaluate(k, data=data): evaluate(v, data=data) for k, v in six.iteritems(statement)}

    elif isinstance(statement, list):
        return [evaluate(item, data=data) for item in statement]

    elif isinstance(statement, six.string_types):
        for name, evaluator in six.iteritems(get_evaluators()):
            if evaluator.has_expressions(statement):
                return evaluator.evaluate(statement, data=data)

    return statement


def extract_vars(statement):
    variables = []

    if isinstance(statement, dict):
        for k, v in six.iteritems(statement):
            variables.extend(extract_vars(k))
            variables.extend(extract_vars(v))

    elif isinstance(statement, list):
        for item in statement:
            variables.extend(extract_vars(item))

    elif isinstance(statement, six.string_types):
        for name, evaluator in six.iteritems(get_evaluators()):
            for var_ref in evaluator.extract_vars(statement):
                for regex_var_extract in evaluator.get_var_extraction_regexes():
                    result = re.search(regex_var_extract, var_ref)
                    var = result.group(1) if result else ""
                    variables.append((evaluator.get_type(), statement, var))

    variables = [v for v in variables if v[2] != ""]

    return sorted(list(set(variables)), key=lambda var: var[2])


def func_has_ctx_arg(func):
    getargspec = (
        inspect.getargspec if six.PY2 else inspect.getfullargspec  # pylint: disable=no-member
    )

    return "context" in getargspec(func).args



# Copyright 2019 Extreme Networks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from orquesta.tests.unit import base as test_base


class JinjaVariableExtractionTest(test_base.ExpressionEvaluatorTest):
    @classmethod
    def setUpClass(cls):
        cls.language = "jinja"
        super(JinjaVariableExtractionTest, cls).setUpClass()

    def test_empty_extraction(self):
        expr = "{{ just_text and $not_a_var and notctx(foo) and notctx(\"bar\") and notctx('fu') }}"

        self.assertListEqual([], self.evaluator.extract_vars(expr))

    def test_single_var_extraction(self):
        expr = "{{ ctx(foobar)  }}"

        expected_vars = ["ctx(foobar)"]

        self.assertListEqual(expected_vars, self.evaluator.extract_vars(expr))

    def test_single_var_extraction_with_single_quotes(self):
        expr = "{{ ctx('foobar')  }}"

        expected_vars = ["ctx('foobar')"]

        self.assertListEqual(expected_vars, self.evaluator.extract_vars(expr))

    def test_single_var_extraction_with_double_quotes(self):
        expr = '{{ ctx("foobar")  }}'

        expected_vars = ['ctx("foobar")']

        self.assertListEqual(expected_vars, self.evaluator.extract_vars(expr))

    def test_single_dotted_var_extraction(self):
        expr = "{{ ctx(foo).bar  }}"

        expected_vars = ["ctx(foo).bar"]

        self.assertListEqual(expected_vars, self.evaluator.extract_vars(expr))

    def test_single_indexing_var_extraction(self):
        expr = "{{ ctx(foo)[0]  }}"

        expected_vars = ["ctx(foo)[0]"]

        self.assertListEqual(expected_vars, self.evaluator.extract_vars(expr))

    def test_single_functional_var_extraction(self):
        expr = "{{ ctx(foo).get(bar)  }}"

        expected_vars = ["ctx(foo).get(bar)"]

        self.assertListEqual(expected_vars, self.evaluator.extract_vars(expr))

    def test_multiple_vars_extraction(self):
        expr = "{{ctx(fubar) ctx(foobar) ctx(foo).get(bar) ctx(fu).bar ctx(foobaz).bar[0]  }}"

        expected_vars = [
            "ctx(foobar)",
            "ctx(foobaz).bar[0]",
            "ctx(foo).get(bar)",
            "ctx(fubar)",
            "ctx(fu).bar",
        ]

        self.assertListEqual(sorted(expected_vars), sorted(self.evaluator.extract_vars(expr)))

    def test_multiple_interleaved_vars_extraction(self):
        expr = "{{ Why the ctx(foobar) are you so ctx(fu).bar serious? }}"

        expected_vars = ["ctx(foobar)", "ctx(fu).bar"]

        self.assertListEqual(expected_vars, sorted(self.evaluator.extract_vars(expr)))



# Copyright 2019 Extreme Networks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from orquesta.tests.unit import base as test_base


class YAQLVariableExtractionTest(test_base.ExpressionEvaluatorTest):
    @classmethod
    def setUpClass(cls):
        cls.language = "yaql"
        super(YAQLVariableExtractionTest, cls).setUpClass()

    def test_empty_extraction(self):
        expr = "<% just_text and $not_a_var and notctx().bar %>"

        self.assertListEqual([], self.evaluator.extract_vars(expr))

    def test_single_var_extraction(self):
        expr = "<% ctx().foobar  %>"

        expected_vars = ["ctx().foobar"]

        self.assertListEqual(expected_vars, self.evaluator.extract_vars(expr))

    def test_single_dotted_var_extraction(self):
        expr = "<% ctx().foo.bar %>"

        expected_vars = ["ctx().foo.bar"]

        self.assertListEqual(expected_vars, self.evaluator.extract_vars(expr))

    def test_single_indexing_var_extraction(self):
        expr = "<% ctx().foo[0]  %>"

        expected_vars = ["ctx().foo[0]"]

        self.assertListEqual(expected_vars, self.evaluator.extract_vars(expr))

    def test_single_functional_var_extraction(self):
        expr = "<% ctx().foo.get(bar)  %>"

        expected_vars = ["ctx().foo.get(bar)"]

        self.assertListEqual(expected_vars, self.evaluator.extract_vars(expr))

    def test_multiple_vars_extraction(self):
        expr = "<%ctx().fubar ctx().foobar ctx().foo.get(bar) ctx().fu.bar ctx().foobaz.bar[0] %>"

        expected_vars = [
            "ctx().foobar",
            "ctx().foobaz.bar[0]",
            "ctx().foo.get(bar)",
            "ctx().fubar",
            "ctx().fu.bar",
        ]

        self.assertListEqual(sorted(expected_vars), sorted(self.evaluator.extract_vars(expr)))

    def test_multiple_interleaved_vars_extraction(self):
        expr = "<% Why the ctx().foobar are you so ctx().fu.bar serious? %>"

        expected_vars = ["ctx().foobar", "ctx().fu.bar"]

        self.assertListEqual(expected_vars, sorted(self.evaluator.extract_vars(expr)))



# Copyright 2019 Extreme Networks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from orquesta import exceptions as exc
from orquesta import graphing
from orquesta.tests.unit import base as test_base
from orquesta.utils import jsonify as json_util


EXPECTED_WF_GRAPH = {
    "directed": True,
    "graph": [],
    "nodes": [
        {"id": "task1"},
        {"id": "task2"},
        {"id": "task3"},
        {"id": "task4"},
        {"id": "task5", "barrier": "*"},
        {"id": "task6"},
        {"id": "task7"},
        {"id": "task8"},
        {"id": "task9"},
    ],
    "adjacency": [
        [
            {"id": "task2", "key": 0, "attr1": "foobar"},
            {"id": "task4", "key": 0},
            {"id": "task7", "key": 0},
            {"id": "task9", "key": 0},
        ],
        [{"id": "task3", "key": 0}],
        [{"id": "task5", "key": 0}],
        [{"id": "task5", "key": 0}],
        [{"id": "task6", "key": 0}],
        [],
        [{"id": "task8", "key": 0}],
        [{"id": "task9", "key": 0}],
        [],
    ],
    "multigraph": True,
}


class WorkflowGraphTest(test_base.WorkflowGraphTest):
    def _add_tasks(self, wf_graph):
        for i in range(1, 10):
            wf_graph.add_task("task" + str(i))

    def _add_transitions(self, wf_graph):
        wf_graph.add_transition("task1", "task2", attr1="foobar")
        wf_graph.add_transition("task2", "task3")
        wf_graph.add_transition("task1", "task4")
        wf_graph.add_transition("task3", "task5")
        wf_graph.add_transition("task4", "task5")
        wf_graph.add_transition("task5", "task6")
        wf_graph.add_transition("task1", "task7")
        wf_graph.add_transition("task7", "task8")
        wf_graph.add_transition("task1", "task9")
        wf_graph.add_transition("task8", "task9")

    def _add_barriers(self, wf_graph):
        wf_graph.update_task("task5", barrier="*")

    def _prep_graph(self):
        wf_graph = graphing.WorkflowGraph()

        self._add_tasks(wf_graph)
        self._add_transitions(wf_graph)
        self._add_barriers(wf_graph)

        return wf_graph

    def test_graph(self):
        wf_graph = self._prep_graph()

        self.assert_graph_equal(wf_graph, EXPECTED_WF_GRAPH)

    def test_graph_roots(self):
        wf_graph = self._prep_graph()

        expected = [{"id": "task1", "name": "task1"}]

        self.assertListEqual(wf_graph.roots, expected)

    def test_graph_leaves(self):
        wf_graph = self._prep_graph()

        expected_leaves = [{"id": "task6", "name": "task6"}, {"id": "task9", "name": "task9"}]

        self.assertListEqual(wf_graph.leaves, expected_leaves)

        # Ensure the underlying graph is not permanently altered.
        expected_roots = [{"id": "task1", "name": "task1"}]

        self.assertListEqual(wf_graph.roots, expected_roots)

    def test_skip_add_tasks(self):
        wf_graph = graphing.WorkflowGraph()

        self._add_transitions(wf_graph)
        self._add_barriers(wf_graph)

        self.assert_graph_equal(wf_graph, EXPECTED_WF_GRAPH)

    def test_duplicate_add_tasks(self):
        wf_graph = self._prep_graph()

        self._add_tasks(wf_graph)

        self.assert_graph_equal(wf_graph, EXPECTED_WF_GRAPH)

    def test_add_same_transition_between_tasks(self):
        wf_graph = self._prep_graph()

        wf_graph.add_transition("task1", "task2", key=0, attr1="foobar")

        self.assert_graph_equal(wf_graph, EXPECTED_WF_GRAPH)

    def test_add_duplicate_transition_between_tasks(self):
        wf_graph = self._prep_graph()

        wf_graph.add_transition("task1", "task2", attr1="fubar")

        expected_wf_graph = json_util.deepcopy(EXPECTED_WF_GRAPH)
        expected_transition = {"id": "task2", "key": 1, "attr1": "fubar"}
        expected_wf_graph["adjacency"][0].append(expected_transition)

        self.assert_graph_equal(wf_graph, expected_wf_graph)

    def test_get_task(self):
        wf_graph = self._prep_graph()

        expected = {"id": "task1"}
        self.assertDictEqual(wf_graph.get_task("task1"), expected)

    def test_get_nonexistent_task(self):
        wf_graph = self._prep_graph()

        self.assertRaises(exc.InvalidTask, wf_graph.get_task, "task999")

    def test_update_task(self):
        wf_graph = self._prep_graph()

        expected = {"id": "task1"}
        self.assertDictEqual(wf_graph.get_task("task1"), expected)

        # Update task with new attributes.
        wf_graph.update_task("task1", attr1="foobar", attr2="fubar")
        expected = {"id": "task1", "attr1": "foobar", "attr2": "fubar"}
        self.assertDictEqual(wf_graph.get_task("task1"), expected)

        # Update an existing attribute.
        wf_graph.update_task("task1", attr2="foobar")
        expected = {"id": "task1", "attr1": "foobar", "attr2": "foobar"}
        self.assertDictEqual(wf_graph.get_task("task1"), expected)

    def test_update_nonexistent_task(self):
        wf_graph = self._prep_graph()

        self.assertRaises(exc.InvalidTask, wf_graph.update_task, "task999", attr1="foobar")

    def test_get_task_attributes(self):
        wf_graph = self._prep_graph()

        wf_graph.update_task("task1", attr1="foobar")
        wf_graph.update_task("task2", attr1="foobar")
        wf_graph.update_task("task4", attr1="foobar")

        expected = {
            "task1": "foobar",
            "task2": "foobar",
            "task3": None,
            "task4": "foobar",
            "task5": None,
            "task6": None,
            "task7": None,
            "task8": None,
            "task9": None,
        }

        self.assertDictEqual(wf_graph.get_task_attributes("attr1"), expected)

    def test_has_transition(self):
        wf_graph = self._prep_graph()

        expected = [("task1", "task2", 0, {"attr1": "foobar"})]
        self.assertListEqual(wf_graph.has_transition("task1", "task2", attr1="foobar"), expected)

        expected = [("task2", "task3", 0, {})]
        self.assertListEqual(wf_graph.has_transition("task2", "task3"), expected)

    def test_get_transition(self):
        wf_graph = self._prep_graph()

        expected = ("task1", "task2", 0, {"attr1": "foobar"})
        self.assertEqual(wf_graph.get_transition("task1", "task2", attr1="foobar"), expected)

        expected = ("task2", "task3", 0, {})
        self.assertEqual(wf_graph.get_transition("task2", "task3"), expected)

    def test_get_nonexistent_transition(self):
        wf_graph = self._prep_graph()

        self.assertRaises(
            exc.InvalidTaskTransition, wf_graph.get_transition, "task1", "task2", attr1="fubar"
        )

        self.assertRaises(exc.InvalidTaskTransition, wf_graph.get_transition, "task998", "task999")

    def test_get_ambiguous_transition(self):
        wf_graph = graphing.WorkflowGraph()

        self._add_tasks(wf_graph)

        wf_graph._graph.add_edge("task1", "task2")
        wf_graph._graph.add_edge("task1", "task2")

        self.assertRaises(exc.AmbiguousTaskTransition, wf_graph.get_transition, "task1", "task2")

    def test_get_next_transitions(self):
        wf_graph = self._prep_graph()

        expected_transitions = [
            ("task1", "task2", 0, {"attr1": "foobar"}),
            ("task1", "task4", 0, {}),
            ("task1", "task7", 0, {}),
            ("task1", "task9", 0, {}),
        ]

        self.assertListEqual(
            sorted(wf_graph.get_next_transitions("task1")), sorted(expected_transitions)
        )

    def test_get_prev_transitions(self):
        wf_graph = self._prep_graph()

        expected_transitions = [("task3", "task5", 0, {}), ("task4", "task5", 0, {})]

        self.assertListEqual(
            sorted(wf_graph.get_prev_transitions("task5")), sorted(expected_transitions)
        )

    def test_task_has_barrier(self):
        wf_graph = self._prep_graph()

        self.assertTrue(wf_graph.has_barrier("task5"))
        self.assertFalse(wf_graph.has_barrier("task9"))

    def test_get_barrier_tasks(self):
        wf_graph = self._prep_graph()

        expected_result = {"task5": {"barrier": "*"}}
        self.assertDictEqual(wf_graph.get_barriers(), expected_result)

    def test_split_from_reused_task(self):
        wf_graph = self._prep_graph()

        self.assertFalse(
            len(wf_graph.get_prev_transitions("task5")) > 1 and not wf_graph.has_barrier("task5")
        )

        self.assertTrue(
            len(wf_graph.get_prev_transitions("task9")) > 1 and not wf_graph.has_barrier("task9")
        )



# Copyright 2019 Extreme Networks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from orquesta import conducting
from orquesta.specs import native as native_specs
from orquesta import statuses
from orquesta.tests.unit import base as test_base


class WorkflowConductorExtendedTaskTest(test_base.WorkflowConductorTest):
    def test_init_task_with_no_action(self):
        wf_def = """
        version: 1.0

        tasks:
          task1:
            next:
              - publish: xyz=123
                do: task2
          task2:
            action: core.noop
        """

        spec = native_specs.WorkflowSpec(wf_def)
        conductor = conducting.WorkflowConductor(spec)
        conductor.request_workflow_status(statuses.RUNNING)

        # Process task1.
        task_name = "task1"
        expected_task_ctx = {}
        self.assert_next_task(conductor, task_name, expected_task_ctx)
        self.forward_task_statuses(conductor, task_name, [statuses.RUNNING, statuses.SUCCEEDED])

        # Process task2.
        task_name = "task2"
        expected_task_ctx = {"xyz": 123}
        self.assert_next_task(conductor, task_name, expected_task_ctx)
        self.forward_task_statuses(conductor, task_name, [statuses.RUNNING, statuses.SUCCEEDED])

        self.assertEqual(conductor.get_workflow_status(), statuses.SUCCEEDED)

    def test_next_task_with_no_action(self):
        wf_def = """
        version: 1.0

        tasks:
          task1:
            action: core.noop
            next:
              - do: task2
          task2:
            next:
              - publish: xyz=123
                do: task3
          task3:
            action: core.noop
        """

        spec = native_specs.WorkflowSpec(wf_def)
        conductor = conducting.WorkflowConductor(spec)
        conductor.request_workflow_status(statuses.RUNNING)

        # Process task1.
        task_name = "task1"
        expected_task_ctx = {}
        self.assert_next_task(conductor, task_name, expected_task_ctx)
        self.forward_task_statuses(conductor, task_name, [statuses.RUNNING, statuses.SUCCEEDED])

        # Process task2.
        task_name = "task2"
        self.assert_next_task(conductor, task_name, expected_task_ctx)
        self.forward_task_statuses(conductor, task_name, [statuses.RUNNING, statuses.SUCCEEDED])

        # Process task3.
        task_name = "task3"
        expected_task_ctx = {"xyz": 123}
        self.assert_next_task(conductor, task_name, expected_task_ctx)
        self.forward_task_statuses(conductor, task_name, [statuses.RUNNING, statuses.SUCCEEDED])

        self.assertEqual(conductor.get_workflow_status(), statuses.SUCCEEDED)



# Copyright 2019 Extreme Networks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from orquesta.tests.unit.composition.native import base


class ErrorHandlingWorkflowComposerTest(base.OrchestraWorkflowComposerTest):
    def test_error_handling_continue(self):
        wf_name = "error-handling-continue"

        expected_wf_graph = {
            "directed": True,
            "graph": {},
            "nodes": [{"id": "task1"}, {"id": "task2"}, {"id": "continue", "splits": ["continue"]}],
            "adjacency": [
                [
                    {"id": "continue", "key": 0, "ref": 1, "criteria": ["<% failed() %>"]},
                    {"id": "task2", "key": 0, "ref": 0, "criteria": ["<% succeeded() %>"]},
                ],
                [{"id": "continue", "key": 0, "ref": 0, "criteria": ["<% succeeded() %>"]}],
                [],
            ],
            "multigraph": True,
        }

        self.assert_compose_to_wf_graph(wf_name, expected_wf_graph)

        expected_wf_ex_graph = {
            "directed": True,
            "graph": {},
            "nodes": [{"id": "task1"}, {"id": "task2"}, {"id": "continue", "splits": ["continue"]}],
            "adjacency": [
                [
                    {"id": "continue", "key": 0, "ref": 1, "criteria": ["<% failed() %>"]},
                    {"id": "task2", "key": 0, "ref": 0, "criteria": ["<% succeeded() %>"]},
                ],
                [{"id": "continue", "key": 0, "ref": 0, "criteria": ["<% succeeded() %>"]}],
                [],
            ],
            "multigraph": True,
        }

        self.assert_compose_to_wf_ex_graph(wf_name, expected_wf_ex_graph)

    def test_error_handling_noop(self):
        wf_name = "error-handling-noop"

        expected_wf_graph = {
            "directed": True,
            "graph": {},
            "nodes": [{"id": "task1"}, {"id": "task2"}, {"id": "continue"}, {"id": "noop"}],
            "adjacency": [
                [
                    {"id": "noop", "key": 0, "ref": 1, "criteria": ["<% failed() %>"]},
                    {"id": "task2", "key": 0, "ref": 0, "criteria": ["<% succeeded() %>"]},
                ],
                [{"id": "continue", "key": 0, "ref": 0, "criteria": ["<% succeeded() %>"]}],
                [],
                [],
            ],
            "multigraph": True,
        }

        self.assert_compose_to_wf_graph(wf_name, expected_wf_graph)

        expected_wf_ex_graph = {
            "directed": True,
            "graph": {},
            "nodes": [{"id": "task1"}, {"id": "task2"}, {"id": "continue"}, {"id": "noop"}],
            "adjacency": [
                [
                    {"id": "noop", "key": 0, "ref": 1, "criteria": ["<% failed() %>"]},
                    {"id": "task2", "key": 0, "ref": 0, "criteria": ["<% succeeded() %>"]},
                ],
                [{"id": "continue", "key": 0, "ref": 0, "criteria": ["<% succeeded() %>"]}],
                [],
                [],
            ],
            "multigraph": True,
        }

        self.assert_compose_to_wf_ex_graph(wf_name, expected_wf_ex_graph)

    def test_error_handling_fail(self):
        wf_name = "error-handling-fail"

        expected_wf_graph = {
            "directed": True,
            "graph": {},
            "nodes": [
                {"id": "task1"},
                {"id": "task2"},
                {"id": "task3"},
                {"id": "continue"},
                {"id": "fail"},
            ],
            "adjacency": [
                [
                    {"id": "task2", "key": 0, "ref": 0, "criteria": ["<% succeeded() %>"]},
                    {"id": "task3", "key": 0, "ref": 1, "criteria": ["<% failed() %>"]},
                ],
                [{"id": "continue", "key": 0, "ref": 0, "criteria": ["<% succeeded() %>"]}],
                [{"id": "fail", "key": 0, "ref": 0, "criteria": ["<% succeeded() %>"]}],
                [],
                [],
            ],
            "multigraph": True,
        }

        self.assert_compose_to_wf_graph(wf_name, expected_wf_graph)

        expected_wf_ex_graph = {
            "directed": True,
            "graph": {},
            "nodes": [
                {"id": "task1"},
                {"id": "task2"},
                {"id": "task3"},
                {"id": "continue"},
                {"id": "fail"},
            ],
            "adjacency": [
                [
                    {"id": "task2", "key": 0, "ref": 0, "criteria": ["<% succeeded() %>"]},
                    {"id": "task3", "key": 0, "ref": 1, "criteria": ["<% failed() %>"]},
                ],
                [{"id": "continue", "key": 0, "ref": 0, "criteria": ["<% succeeded() %>"]}],
                [{"id": "fail", "key": 0, "ref": 0, "criteria": ["<% succeeded() %>"]}],
                [],
                [],
            ],
            "multigraph": True,
        }

        self.assert_compose_to_wf_ex_graph(wf_name, expected_wf_ex_graph)



# Copyright 2019 Extreme Networks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import threading

from stevedore import extension


_EXP_FUNC_CATALOG = None
_EXP_FUNC_CATALOG_LOCK = threading.Lock()


def load():
    global _EXP_FUNC_CATALOG
    global _EXP_FUNC_CATALOG_LOCK

    with _EXP_FUNC_CATALOG_LOCK:
        if _EXP_FUNC_CATALOG is None:
            _EXP_FUNC_CATALOG = {}

            mgr = extension.ExtensionManager(
                namespace="orquesta.expressions.functions", invoke_on_load=False
            )

            for name in mgr.names():
                _EXP_FUNC_CATALOG[name] = mgr[name].plugin

    return _EXP_FUNC_CATALOG



# Copyright 2019 Extreme Networks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from orquesta.tests.unit import base as test_base


class YAQLFacadeValidationTest(test_base.ExpressionFacadeEvaluatorTest):
    def test_basic_validate(self):
        self.assertListEqual([], self.validate(None))
        self.assertListEqual([], self.validate("<% 1 %>"))
        self.assertListEqual([], self.validate("<% abc %>"))
        self.assertListEqual([], self.validate("<% 1 + 2 %>"))
        self.assertListEqual([], self.validate("<% ctx().foo %>"))
        self.assertListEqual([], self.validate("<% ctx(foo) %>"))
        self.assertListEqual([], self.validate("<% ctx().a1 + ctx(a2) %>"))

    def test_parse_error(self):
        expr = "<% <% ctx().foo %> %>"
        errors = self.validate(expr)

        self.assertEqual(1, len(errors))
        self.assertIn("Parse error", errors[0]["message"])

    def test_lexical_error(self):
        expr = '<% {"a": 123} %>'
        errors = self.validate(expr)

        self.assertEqual(1, len(errors))
        self.assertIn("Lexical error", errors[0]["message"])

    def test_multiple_errors(self):
        expr = '<% 1 +/ 2 %> and <% {"a": 123} %>'
        errors = self.validate(expr)

        self.assertEqual(2, len(errors))
        self.assertIn("Parse error", errors[0]["message"])
        self.assertIn("Lexical error", errors[1]["message"])

    def test_list_validate(self):
        self.assertListEqual([], self.validate([]))

        target = ["<% 1 %>", "xyz", 123, True]
        self.assertListEqual([], self.validate(target))

        target = ["<% 1 %>", ["<% a %>", "x"]]
        self.assertListEqual([], self.validate(target))

        target = ["<% 1 %>", ["<% a %>", "x"], {"k1": "<% 3 %>"}]
        self.assertListEqual([], self.validate(target))

    def test_list_multiple_errors(self):
        target = ["<% 1 +/ 2 %>", ["<% <% ctx().foo %>"], {"k1": '<% {"a": 1} %>'}]
        result = self.validate(target)
        self.assertEqual(3, len(result))

    def test_dict_validate(self):
        self.assertListEqual([], self.validate({}))

        target = {
            "k1": "<% 1 %>",
            "k2": "foobar",
            "<% k3 %>": 789,
            "k4": ["<% abc %>", "xyz", 123, False, {"k": "v"}],
            "depth-1": {
                "depth-1-1": {
                    "depth-1-1-2": "<% ctx().a1 + ctx(a2) %>",
                    "depth-1-1-3": ["<% ctx().foobar %>", "xyz"],
                }
            },
        }

        self.assertListEqual([], self.validate(target))

    def test_dict_multiple_errors(self):
        target = {
            "k1": "<% 1 +/ 2 %>",
            "k2": "foobar",
            "<% <% ctx().foo %>": 789,
            "k4": ["<% abc %>", "xyz", 123, False, {"k": "v"}],
            "depth-1": {
                "depth-1-1": {
                    "depth-1-1-2": '<% {"a": 1} %>',
                    "depth-1-1-3": ["<% ctx().foobar %>", "<% 3 +/ 4 %>"],
                }
            },
        }

        result = self.validate(target)
        self.assertEqual(4, len(result))



# Copyright 2019 Extreme Networks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging

import chardet
import six

if six.PY2:
    import __builtin__  # pylint: disable=import-error


LOG = logging.getLogger(__name__)


def unescape(s):
    return s.decode("string_escape") if six.PY2 else bytes(s, "utf-8").decode("unicode_escape")


def encoding(s):
    return chardet.detect(s)["encoding"]


def unicode(s, force=False, encoding_type=None):
    if not isinstance(s, six.string_types) and not force:
        return s

    if six.PY2:
        if not force and (isinstance(s, __builtin__.unicode) or encoding(s) == "utf-8"):
            return s

        return __builtin__.unicode(s, encoding_type) if encoding_type else __builtin__.unicode(s)

    return str(s)



# Copyright 2019 Extreme Networks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from orquesta.tests.unit.specs.mistral import base as test_base


class WorkflowSpecValidationTest(test_base.MistralWorkflowSpecTest):
    def test_success(self):
        wf_def = """
            version: '2.0'

            sequential:
                description: A basic sequential workflow.
                tasks:
                    task1:
                        action: std.noop
                        on-success:
                            - task2
                    task2:
                        action: std.noop
                        on-success:
                            - task3
                    task3:
                        action: std.noop
        """

        wf_spec = self.instantiate(wf_def)

        self.assertDictEqual(wf_spec.inspect(), {})

    def test_on_success_conditional(self):
        wf_def = """
            version: '2.0'

            sequential:
                description: A basic sequential workflow.
                vars:
                  xxx: true
                tasks:
                    task1:
                        action: std.noop
                        on-success:
                            - task2: "{{ _.xxx }}"
                    task2:
                        action: std.noop
        """

        wf_spec = self.instantiate(wf_def)

        self.assertDictEqual(wf_spec.inspect(), {})

    def test_on_error_conditional(self):
        wf_def = """
            version: '2.0'

            sequential:
                description: A basic sequential workflow.
                vars:
                  xxx: true
                tasks:
                    task1:
                        action: std.noop
                        on-error:
                            - task2: "{{ _.xxx }}"
                    task2:
                        action: std.noop
        """

        wf_spec = self.instantiate(wf_def)

        self.assertDictEqual(wf_spec.inspect(), {})

    def test_on_complete_conditional(self):
        wf_def = """
            version: '2.0'

            sequential:
                description: A basic sequential workflow.
                vars:
                  xxx: true
                tasks:
                    task1:
                        action: std.noop
                        on-complete:
                            - task2: "{{ _.xxx }}"
                    task2:
                        action: std.noop
        """

        wf_spec = self.instantiate(wf_def)

        self.assertDictEqual(wf_spec.inspect(), {})

    def test_missing_task_list(self):
        wf_def = """
            version: '2.0'

            sequential:
                description: A basic sequential workflow.
        """

        wf_spec = self.instantiate(wf_def)

        expected_errors = {
            "syntax": [
                {
                    "message": "'tasks' is a required property",
                    "schema_path": "required",
                    "spec_path": None,
                }
            ]
        }

        self.assertDictEqual(wf_spec.inspect(), expected_errors)

    def test_empty_task_list(self):
        wf_def = """
            version: '2.0'

            sequential:
                description: A basic sequential workflow.
                tasks: {}
        """

        wf_spec = self.instantiate(wf_def)

        expected_errors = {
            "syntax": [
                {
                    "message": "{} does not have enough properties",
                    "schema_path": "properties.tasks.minProperties",
                    "spec_path": "tasks",
                }
            ]
        }

        self.assertDictEqual(wf_spec.inspect(), expected_errors)

    def test_task_default_list(self):
        wf_def = """
            version: '2.0'

            sequential:
                description: A basic sequential workflow.
                task-defaults:
                   concurrency: 1
                   keep-result: true
                   retry:
                       delay: 5
                       count: 10
                   safe-rerun: true
                   wait-before: 3
                   wait-after: 9
                   pause-before: false
                   target: some_node
                   timeout: 90
                   on-success:
                       - task2
                   on-error:
                       - task2
                   on-complete:
                       - task3
                tasks:
                    task1:
                        action: std.noop
                    task2:
                        action: std.noop
                    task3:
                        action: std.noop
        """

        wf_spec = self.instantiate(wf_def)

        self.assertDictEqual(wf_spec.inspect(), {})

    def test_empty_task_default_list(self):
        wf_def = """
            version: '2.0'

            sequential:
                description: A basic sequential workflow.
                task-defaults: {}
                tasks:
                    task1:
                        action: std.noop
        """

        wf_spec = self.instantiate(wf_def)

        self.assertDictEqual(wf_spec.inspect(), {})

    def test_workflow_type_direct(self):
        wf_def = """
            version: '2.0'

            sequential:
                description: A basic sequential workflow.
                type: direct
                tasks:
                    task1:
                        action: std.noop
        """

        wf_spec = self.instantiate(wf_def)

        self.assertDictEqual(wf_spec.inspect(), {})

    def test_workflow_type_reverse(self):
        wf_def = """
            version: '2.0'

            sequential:
                description: A basic sequential workflow.
                type: reverse
                tasks:
                    task1:
                        action: std.noop
        """

        wf_spec = self.instantiate(wf_def)

        self.assertDictEqual(wf_spec.inspect(), {})

    def test_bad_workflow_type(self):
        wf_def = """
            version: '2.0'

            sequential:
                description: A basic sequential workflow.
                type: junk
                tasks:
                    task1:
                        action: std.noop
        """

        wf_spec = self.instantiate(wf_def)

        expected_errors = {
            "syntax": [
                {
                    "message": "'junk' is not one of ['reverse', 'direct']",
                    "schema_path": "properties.type.enum",
                    "spec_path": "type",
                }
            ]
        }

        self.assertDictEqual(wf_spec.inspect(), expected_errors)

    def test_workflow_with_output_on_error(self):
        wf_def = """
            version: '2.0'

            sequential:
                description: A basic sequential workflow.
                vars:
                    jinja_expr: "abc"
                    test_jinja: "xyz"
                output-on-error:
                    stdout: "abc"
                    jinja_expr: "{{ _.test_jinja }}"
                    yaql_expr: <% $.test_yaql %>

                tasks:
                    task1:
                        action: std.noop
        """

        wf_spec = self.instantiate(wf_def)

        self.assertDictEqual(wf_spec.inspect(), {})

    def test_empty_output_on_error(self):
        wf_def = """
            version: '2.0'

            sequential:
                description: A basic sequential workflow.
                output-on-error: {}
                tasks:
                    task1:
                        action: std.noop
        """

        wf_spec = self.instantiate(wf_def)

        expected_errors = {
            "syntax": [
                {
                    "message": "{} does not have enough properties",
                    "schema_path": "properties.output-on-error.minProperties",
                    "spec_path": "output-on-error",
                }
            ]
        }

        self.assertDictEqual(wf_spec.inspect(), expected_errors)

    def test_task_policies(self):
        wf_def = """
            version: '2.0'

            sequential:
                description: A basic sequential workflow.

                tasks:
                    task1:
                        action: std.noop
                        concurrency: 1
                        keep-result: true
                        retry:
                            delay: 5
                            count: 10
                        safe-rerun: true
                        wait-before: 3
                        wait-after: 9
                        pause-before: false
                        target: some_node
                        timeout: 90
        """

        wf_spec = self.instantiate(wf_def)

        self.assertDictEqual(wf_spec.inspect(), {})

    def test_task_policies_with_expressions(self):
        wf_def = """
            version: '2.0'

            sequential:
                description: A basic sequential workflow.

                tasks:
                    task1:
                        action: std.noop
                        concurrency: "{{ 3 | int }}"
                        keep-result: "{{ True }}"
                        retry:
                            delay: "{{ 5 | int }}"
                            count: "{{ 10 | int }}"
                        safe-rerun: "{{ True }}"
                        wait-before: "{{ 3 | int }}"
                        wait-after: "{{ 9 | int }}"
                        pause-before: "{{ False }}"
                        target: "{{ 'some_node' }}"
                        timeout: "{{ 39 | int }}"
        """

        wf_spec = self.instantiate(wf_def)

        self.assertDictEqual(wf_spec.inspect(), {})

    def test_task_policies_validation_errors(self):
        wf_def = """
            version: '2.0'

            sequential:
                description: A basic sequential workflow.

                tasks:
                    task1:
                        action: std.noop
                        concurrency: []
                        keep-result: {}
                        retry: ''
                        safe-rerun: []
                        wait-before: []
                        wait-after: []
                        pause-before: []
                        target: {}
                        timeout: []
        """

        wf_spec = self.instantiate(wf_def)

        expected_errors = {
            "syntax": [
                {
                    "message": "[] is not valid under any of the given schemas",
                    "schema_path": (
                        "properties.tasks.patternProperties" ".^\\w+$.properties.concurrency.oneOf"
                    ),
                    "spec_path": "tasks.task1.concurrency",
                },
                {
                    "message": "{} is not valid under any of the given schemas",
                    "schema_path": (
                        "properties.tasks.patternProperties" ".^\\w+$.properties.keep-result.oneOf"
                    ),
                    "spec_path": "tasks.task1.keep-result",
                },
                {
                    "message": "[] is not valid under any of the given schemas",
                    "schema_path": (
                        "properties.tasks.patternProperties" ".^\\w+$.properties.pause-before.oneOf"
                    ),
                    "spec_path": "tasks.task1.pause-before",
                },
                {
                    "message": "'' is not of type 'object'",
                    "schema_path": (
                        "properties.tasks.patternProperties" ".^\\w+$.properties.retry.type"
                    ),
                    "spec_path": "tasks.task1.retry",
                },
                {
                    "message": "[] is not valid under any of the given schemas",
                    "schema_path": (
                        "properties.tasks.patternProperties" ".^\\w+$.properties.safe-rerun.oneOf"
                    ),
                    "spec_path": "tasks.task1.safe-rerun",
                },
                {
                    "message": "{} is not of type 'string'",
                    "schema_path": (
                        "properties.tasks.patternProperties" ".^\\w+$.properties.target.type"
                    ),
                    "spec_path": "tasks.task1.target",
                },
                {
                    "message": "[] is not valid under any of the given schemas",
                    "schema_path": (
                        "properties.tasks.patternProperties" ".^\\w+$.properties.timeout.oneOf"
                    ),
                    "spec_path": "tasks.task1.timeout",
                },
                {
                    "message": "[] is not valid under any of the given schemas",
                    "schema_path": (
                        "properties.tasks.patternProperties" ".^\\w+$.properties.wait-after.oneOf"
                    ),
                    "spec_path": "tasks.task1.wait-after",
                },
                {
                    "message": "[] is not valid under any of the given schemas",
                    "schema_path": (
                        "properties.tasks.patternProperties" ".^\\w+$.properties.wait-before.oneOf"
                    ),
                    "spec_path": "tasks.task1.wait-before",
                },
            ]
        }

        self.assertDictEqual(wf_spec.inspect(), expected_errors)

    def test_task_publish_on_error(self):
        wf_def = """
            version: '2.0'

            sequential:
                description: A basic sequential workflow.

                tasks:
                    task1:
                        action: std.noop
                        publish-on-error:
                            stdout: "{{ 'abc' }}"
                            stderr:  <$ 'xyz' %>
        """

        wf_spec = self.instantiate(wf_def)

        self.assertDictEqual(wf_spec.inspect(), {})
